import typing, clr, abc
from System import ICloneable, Array_1, DateTime, DayOfWeek, ReadOnlySpan_1, Span_1, IFormatProvider, ArgumentException, Exception, TimeSpan, StringComparer, IEquatable_1, Guid
from System.Runtime.Serialization import IDeserializationCallback, SerializationInfo, StreamingContext
from System.Reflection import Assembly, MethodBase
from System.Text import Rune
from System.Collections import IDictionary, IEnumerator

class Calendar(ICloneable, abc.ABC):
    CurrentEra : int
    @property
    def AlgorithmType(self) -> CalendarAlgorithmType: ...
    @property
    def Eras(self) -> Array_1[int]: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def MaxSupportedDateTime(self) -> DateTime: ...
    @property
    def MinSupportedDateTime(self) -> DateTime: ...
    @property
    def TwoDigitYearMax(self) -> int: ...
    @TwoDigitYearMax.setter
    def TwoDigitYearMax(self, value: int) -> int: ...
    def AddDays(self, time: DateTime, days: int) -> DateTime: ...
    def AddHours(self, time: DateTime, hours: int) -> DateTime: ...
    def AddMilliseconds(self, time: DateTime, milliseconds: float) -> DateTime: ...
    def AddMinutes(self, time: DateTime, minutes: int) -> DateTime: ...
    @abc.abstractmethod
    def AddMonths(self, time: DateTime, months: int) -> DateTime: ...
    def AddSeconds(self, time: DateTime, seconds: int) -> DateTime: ...
    def AddWeeks(self, time: DateTime, weeks: int) -> DateTime: ...
    @abc.abstractmethod
    def AddYears(self, time: DateTime, years: int) -> DateTime: ...
    def Clone(self) -> typing.Any: ...
    @abc.abstractmethod
    def GetDayOfMonth(self, time: DateTime) -> int: ...
    @abc.abstractmethod
    def GetDayOfWeek(self, time: DateTime) -> DayOfWeek: ...
    @abc.abstractmethod
    def GetDayOfYear(self, time: DateTime) -> int: ...
    @abc.abstractmethod
    def GetEra(self, time: DateTime) -> int: ...
    def GetHour(self, time: DateTime) -> int: ...
    def GetMilliseconds(self, time: DateTime) -> float: ...
    def GetMinute(self, time: DateTime) -> int: ...
    @abc.abstractmethod
    def GetMonth(self, time: DateTime) -> int: ...
    def GetSecond(self, time: DateTime) -> int: ...
    def GetWeekOfYear(self, time: DateTime, rule: CalendarWeekRule, firstDayOfWeek: DayOfWeek) -> int: ...
    @abc.abstractmethod
    def GetYear(self, time: DateTime) -> int: ...
    @staticmethod
    def ReadOnly(calendar: Calendar) -> Calendar: ...
    def ToFourDigitYear(self, year: int) -> int: ...
    # Skipped GetDaysInMonth due to it being static, abstract and generic.

    GetDaysInMonth : GetDaysInMonth_MethodGroup
    class GetDaysInMonth_MethodGroup:
        @typing.overload
        def __call__(self, year: int, month: int) -> int:...
        @typing.overload
        def __call__(self, year: int, month: int, era: int) -> int:...

    # Skipped GetDaysInYear due to it being static, abstract and generic.

    GetDaysInYear : GetDaysInYear_MethodGroup
    class GetDaysInYear_MethodGroup:
        @typing.overload
        def __call__(self, year: int) -> int:...
        @typing.overload
        def __call__(self, year: int, era: int) -> int:...

    # Skipped GetLeapMonth due to it being static, abstract and generic.

    GetLeapMonth : GetLeapMonth_MethodGroup
    class GetLeapMonth_MethodGroup:
        @typing.overload
        def __call__(self, year: int) -> int:...
        @typing.overload
        def __call__(self, year: int, era: int) -> int:...

    # Skipped GetMonthsInYear due to it being static, abstract and generic.

    GetMonthsInYear : GetMonthsInYear_MethodGroup
    class GetMonthsInYear_MethodGroup:
        @typing.overload
        def __call__(self, year: int) -> int:...
        @typing.overload
        def __call__(self, year: int, era: int) -> int:...

    # Skipped IsLeapDay due to it being static, abstract and generic.

    IsLeapDay : IsLeapDay_MethodGroup
    class IsLeapDay_MethodGroup:
        @typing.overload
        def __call__(self, year: int, month: int, day: int) -> bool:...
        @typing.overload
        def __call__(self, year: int, month: int, day: int, era: int) -> bool:...

    # Skipped IsLeapMonth due to it being static, abstract and generic.

    IsLeapMonth : IsLeapMonth_MethodGroup
    class IsLeapMonth_MethodGroup:
        @typing.overload
        def __call__(self, year: int, month: int) -> bool:...
        @typing.overload
        def __call__(self, year: int, month: int, era: int) -> bool:...

    # Skipped IsLeapYear due to it being static, abstract and generic.

    IsLeapYear : IsLeapYear_MethodGroup
    class IsLeapYear_MethodGroup:
        @typing.overload
        def __call__(self, year: int) -> bool:...
        @typing.overload
        def __call__(self, year: int, era: int) -> bool:...

    # Skipped ToDateTime due to it being static, abstract and generic.

    ToDateTime : ToDateTime_MethodGroup
    class ToDateTime_MethodGroup:
        @typing.overload
        def __call__(self, year: int, month: int, day: int, hour: int, minute: int, second: int, millisecond: int) -> DateTime:...
        @typing.overload
        def __call__(self, year: int, month: int, day: int, hour: int, minute: int, second: int, millisecond: int, era: int) -> DateTime:...



class CalendarAlgorithmType(typing.SupportsInt):
    @typing.overload
    def __init__(self, value : int) -> None: ...
    @typing.overload
    def __init__(self, value : int, force_if_true: bool) -> None: ...
    def __int__(self) -> int: ...
    
    # Values:
    Unknown : CalendarAlgorithmType # 0
    SolarCalendar : CalendarAlgorithmType # 1
    LunarCalendar : CalendarAlgorithmType # 2
    LunisolarCalendar : CalendarAlgorithmType # 3


class CalendarWeekRule(typing.SupportsInt):
    @typing.overload
    def __init__(self, value : int) -> None: ...
    @typing.overload
    def __init__(self, value : int, force_if_true: bool) -> None: ...
    def __int__(self) -> int: ...
    
    # Values:
    FirstDay : CalendarWeekRule # 0
    FirstFullWeek : CalendarWeekRule # 1
    FirstFourDayWeek : CalendarWeekRule # 2


class CharUnicodeInfo(abc.ABC):
    # Skipped GetDecimalDigitValue due to it being static, abstract and generic.

    GetDecimalDigitValue : GetDecimalDigitValue_MethodGroup
    class GetDecimalDigitValue_MethodGroup:
        @typing.overload
        def __call__(self, ch: str) -> int:...
        @typing.overload
        def __call__(self, s: str, index: int) -> int:...

    # Skipped GetDigitValue due to it being static, abstract and generic.

    GetDigitValue : GetDigitValue_MethodGroup
    class GetDigitValue_MethodGroup:
        @typing.overload
        def __call__(self, ch: str) -> int:...
        @typing.overload
        def __call__(self, s: str, index: int) -> int:...

    # Skipped GetNumericValue due to it being static, abstract and generic.

    GetNumericValue : GetNumericValue_MethodGroup
    class GetNumericValue_MethodGroup:
        @typing.overload
        def __call__(self, ch: str) -> float:...
        @typing.overload
        def __call__(self, s: str, index: int) -> float:...

    # Skipped GetUnicodeCategory due to it being static, abstract and generic.

    GetUnicodeCategory : GetUnicodeCategory_MethodGroup
    class GetUnicodeCategory_MethodGroup:
        @typing.overload
        def __call__(self, ch: str) -> UnicodeCategory:...
        @typing.overload
        def __call__(self, codePoint: int) -> UnicodeCategory:...
        @typing.overload
        def __call__(self, s: str, index: int) -> UnicodeCategory:...



class ChineseLunisolarCalendar(EastAsianLunisolarCalendar):
    def __init__(self) -> None: ...
    ChineseEra : int
    @property
    def AlgorithmType(self) -> CalendarAlgorithmType: ...
    @property
    def Eras(self) -> Array_1[int]: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def MaxSupportedDateTime(self) -> DateTime: ...
    @property
    def MinSupportedDateTime(self) -> DateTime: ...
    @property
    def TwoDigitYearMax(self) -> int: ...
    @TwoDigitYearMax.setter
    def TwoDigitYearMax(self, value: int) -> int: ...
    def GetEra(self, time: DateTime) -> int: ...


class CompareInfo(IDeserializationCallback):
    @property
    def LCID(self) -> int: ...
    @property
    def Name(self) -> str: ...
    @property
    def Version(self) -> SortVersion: ...
    def Equals(self, value: typing.Any) -> bool: ...
    def GetSortKeyLength(self, source: ReadOnlySpan_1[str], options: CompareOptions = ...) -> int: ...
    def ToString(self) -> str: ...
    # Skipped Compare due to it being static, abstract and generic.

    Compare : Compare_MethodGroup
    class Compare_MethodGroup:
        @typing.overload
        def __call__(self, string1: str, string2: str) -> int:...
        @typing.overload
        def __call__(self, string1: ReadOnlySpan_1[str], string2: ReadOnlySpan_1[str], options: CompareOptions = ...) -> int:...
        @typing.overload
        def __call__(self, string1: str, string2: str, options: CompareOptions) -> int:...
        @typing.overload
        def __call__(self, string1: str, offset1: int, string2: str, offset2: int) -> int:...
        @typing.overload
        def __call__(self, string1: str, offset1: int, string2: str, offset2: int, options: CompareOptions) -> int:...
        @typing.overload
        def __call__(self, string1: str, offset1: int, length1: int, string2: str, offset2: int, length2: int) -> int:...
        @typing.overload
        def __call__(self, string1: str, offset1: int, length1: int, string2: str, offset2: int, length2: int, options: CompareOptions) -> int:...

    # Skipped GetCompareInfo due to it being static, abstract and generic.

    GetCompareInfo : GetCompareInfo_MethodGroup
    class GetCompareInfo_MethodGroup:
        @typing.overload
        def __call__(self, culture: int) -> CompareInfo:...
        @typing.overload
        def __call__(self, name: str) -> CompareInfo:...
        @typing.overload
        def __call__(self, culture: int, assembly: Assembly) -> CompareInfo:...
        @typing.overload
        def __call__(self, name: str, assembly: Assembly) -> CompareInfo:...

    # Skipped GetHashCode due to it being static, abstract and generic.

    GetHashCode : GetHashCode_MethodGroup
    class GetHashCode_MethodGroup:
        @typing.overload
        def __call__(self) -> int:...
        @typing.overload
        def __call__(self, source: ReadOnlySpan_1[str], options: CompareOptions) -> int:...
        @typing.overload
        def __call__(self, source: str, options: CompareOptions) -> int:...

    # Skipped GetSortKey due to it being static, abstract and generic.

    GetSortKey : GetSortKey_MethodGroup
    class GetSortKey_MethodGroup:
        @typing.overload
        def __call__(self, source: str) -> SortKey:...
        @typing.overload
        def __call__(self, source: str, options: CompareOptions) -> SortKey:...
        @typing.overload
        def __call__(self, source: ReadOnlySpan_1[str], destination: Span_1[int], options: CompareOptions = ...) -> int:...

    # Skipped IndexOf due to it being static, abstract and generic.

    IndexOf : IndexOf_MethodGroup
    class IndexOf_MethodGroup:
        @typing.overload
        def __call__(self, source: str, value: str) -> int:...
        # Method IndexOf(source : String, value : String) was skipped since it collides with above method
        @typing.overload
        def __call__(self, source: ReadOnlySpan_1[str], value: ReadOnlySpan_1[str], options: CompareOptions = ...) -> int:...
        @typing.overload
        def __call__(self, source: ReadOnlySpan_1[str], value: Rune, options: CompareOptions = ...) -> int:...
        @typing.overload
        def __call__(self, source: str, value: str, startIndex: int) -> int:...
        # Method IndexOf(source : String, value : String, startIndex : Int32) was skipped since it collides with above method
        @typing.overload
        def __call__(self, source: str, value: str, options: CompareOptions) -> int:...
        # Method IndexOf(source : String, value : String, options : CompareOptions) was skipped since it collides with above method
        @typing.overload
        def __call__(self, source: ReadOnlySpan_1[str], value: ReadOnlySpan_1[str], options: CompareOptions, matchLength: clr.Reference[int]) -> int:...
        @typing.overload
        def __call__(self, source: str, value: str, startIndex: int, count: int) -> int:...
        # Method IndexOf(source : String, value : String, startIndex : Int32, count : Int32) was skipped since it collides with above method
        @typing.overload
        def __call__(self, source: str, value: str, startIndex: int, options: CompareOptions) -> int:...
        # Method IndexOf(source : String, value : String, startIndex : Int32, options : CompareOptions) was skipped since it collides with above method
        @typing.overload
        def __call__(self, source: str, value: str, startIndex: int, count: int, options: CompareOptions) -> int:...
        # Method IndexOf(source : String, value : String, startIndex : Int32, count : Int32, options : CompareOptions) was skipped since it collides with above method

    # Skipped IsPrefix due to it being static, abstract and generic.

    IsPrefix : IsPrefix_MethodGroup
    class IsPrefix_MethodGroup:
        @typing.overload
        def __call__(self, source: str, prefix: str) -> bool:...
        @typing.overload
        def __call__(self, source: ReadOnlySpan_1[str], prefix: ReadOnlySpan_1[str], options: CompareOptions = ...) -> bool:...
        @typing.overload
        def __call__(self, source: str, prefix: str, options: CompareOptions) -> bool:...
        @typing.overload
        def __call__(self, source: ReadOnlySpan_1[str], prefix: ReadOnlySpan_1[str], options: CompareOptions, matchLength: clr.Reference[int]) -> bool:...

    # Skipped IsSortable due to it being static, abstract and generic.

    IsSortable : IsSortable_MethodGroup
    class IsSortable_MethodGroup:
        @typing.overload
        def __call__(self, ch: str) -> bool:...
        @typing.overload
        def __call__(self, text: ReadOnlySpan_1[str]) -> bool:...
        # Method IsSortable(text : String) was skipped since it collides with above method
        @typing.overload
        def __call__(self, value: Rune) -> bool:...

    # Skipped IsSuffix due to it being static, abstract and generic.

    IsSuffix : IsSuffix_MethodGroup
    class IsSuffix_MethodGroup:
        @typing.overload
        def __call__(self, source: str, suffix: str) -> bool:...
        @typing.overload
        def __call__(self, source: ReadOnlySpan_1[str], suffix: ReadOnlySpan_1[str], options: CompareOptions = ...) -> bool:...
        @typing.overload
        def __call__(self, source: str, suffix: str, options: CompareOptions) -> bool:...
        @typing.overload
        def __call__(self, source: ReadOnlySpan_1[str], suffix: ReadOnlySpan_1[str], options: CompareOptions, matchLength: clr.Reference[int]) -> bool:...

    # Skipped LastIndexOf due to it being static, abstract and generic.

    LastIndexOf : LastIndexOf_MethodGroup
    class LastIndexOf_MethodGroup:
        @typing.overload
        def __call__(self, source: str, value: str) -> int:...
        # Method LastIndexOf(source : String, value : String) was skipped since it collides with above method
        @typing.overload
        def __call__(self, source: ReadOnlySpan_1[str], value: ReadOnlySpan_1[str], options: CompareOptions = ...) -> int:...
        @typing.overload
        def __call__(self, source: ReadOnlySpan_1[str], value: Rune, options: CompareOptions = ...) -> int:...
        @typing.overload
        def __call__(self, source: str, value: str, startIndex: int) -> int:...
        # Method LastIndexOf(source : String, value : String, startIndex : Int32) was skipped since it collides with above method
        @typing.overload
        def __call__(self, source: str, value: str, options: CompareOptions) -> int:...
        # Method LastIndexOf(source : String, value : String, options : CompareOptions) was skipped since it collides with above method
        @typing.overload
        def __call__(self, source: ReadOnlySpan_1[str], value: ReadOnlySpan_1[str], options: CompareOptions, matchLength: clr.Reference[int]) -> int:...
        @typing.overload
        def __call__(self, source: str, value: str, startIndex: int, count: int) -> int:...
        # Method LastIndexOf(source : String, value : String, startIndex : Int32, count : Int32) was skipped since it collides with above method
        @typing.overload
        def __call__(self, source: str, value: str, startIndex: int, options: CompareOptions) -> int:...
        # Method LastIndexOf(source : String, value : String, startIndex : Int32, options : CompareOptions) was skipped since it collides with above method
        @typing.overload
        def __call__(self, source: str, value: str, startIndex: int, count: int, options: CompareOptions) -> int:...
        # Method LastIndexOf(source : String, value : String, startIndex : Int32, count : Int32, options : CompareOptions) was skipped since it collides with above method



class CompareOptions(typing.SupportsInt):
    @typing.overload
    def __init__(self, value : int) -> None: ...
    @typing.overload
    def __init__(self, value : int, force_if_true: bool) -> None: ...
    def __int__(self) -> int: ...
    
    # Values:
    None_ : CompareOptions # 0
    IgnoreCase : CompareOptions # 1
    IgnoreNonSpace : CompareOptions # 2
    IgnoreSymbols : CompareOptions # 4
    IgnoreKanaType : CompareOptions # 8
    IgnoreWidth : CompareOptions # 16
    OrdinalIgnoreCase : CompareOptions # 268435456
    StringSort : CompareOptions # 536870912
    Ordinal : CompareOptions # 1073741824


class CultureInfo(ICloneable, IFormatProvider):
    @typing.overload
    def __init__(self, culture: int) -> None: ...
    @typing.overload
    def __init__(self, culture: int, useUserOverride: bool) -> None: ...
    @typing.overload
    def __init__(self, name: str) -> None: ...
    @typing.overload
    def __init__(self, name: str, useUserOverride: bool) -> None: ...
    @property
    def Calendar(self) -> Calendar: ...
    @property
    def CompareInfo(self) -> CompareInfo: ...
    @property
    def CultureTypes(self) -> CultureTypes: ...
    @classmethod
    @property
    def CurrentCulture(cls) -> CultureInfo: ...
    @classmethod
    @CurrentCulture.setter
    def CurrentCulture(cls, value: CultureInfo) -> CultureInfo: ...
    @classmethod
    @property
    def CurrentUICulture(cls) -> CultureInfo: ...
    @classmethod
    @CurrentUICulture.setter
    def CurrentUICulture(cls, value: CultureInfo) -> CultureInfo: ...
    @property
    def DateTimeFormat(self) -> DateTimeFormatInfo: ...
    @DateTimeFormat.setter
    def DateTimeFormat(self, value: DateTimeFormatInfo) -> DateTimeFormatInfo: ...
    @classmethod
    @property
    def DefaultThreadCurrentCulture(cls) -> CultureInfo: ...
    @classmethod
    @DefaultThreadCurrentCulture.setter
    def DefaultThreadCurrentCulture(cls, value: CultureInfo) -> CultureInfo: ...
    @classmethod
    @property
    def DefaultThreadCurrentUICulture(cls) -> CultureInfo: ...
    @classmethod
    @DefaultThreadCurrentUICulture.setter
    def DefaultThreadCurrentUICulture(cls, value: CultureInfo) -> CultureInfo: ...
    @property
    def DisplayName(self) -> str: ...
    @property
    def EnglishName(self) -> str: ...
    @property
    def IetfLanguageTag(self) -> str: ...
    @classmethod
    @property
    def InstalledUICulture(cls) -> CultureInfo: ...
    @classmethod
    @property
    def InvariantCulture(cls) -> CultureInfo: ...
    @property
    def IsNeutralCulture(self) -> bool: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def KeyboardLayoutId(self) -> int: ...
    @property
    def LCID(self) -> int: ...
    @property
    def Name(self) -> str: ...
    @property
    def NativeName(self) -> str: ...
    @property
    def NumberFormat(self) -> NumberFormatInfo: ...
    @NumberFormat.setter
    def NumberFormat(self, value: NumberFormatInfo) -> NumberFormatInfo: ...
    @property
    def OptionalCalendars(self) -> Array_1[Calendar]: ...
    @property
    def Parent(self) -> CultureInfo: ...
    @property
    def TextInfo(self) -> TextInfo: ...
    @property
    def ThreeLetterISOLanguageName(self) -> str: ...
    @property
    def ThreeLetterWindowsLanguageName(self) -> str: ...
    @property
    def TwoLetterISOLanguageName(self) -> str: ...
    @property
    def UseUserOverride(self) -> bool: ...
    def ClearCachedData(self) -> None: ...
    def Clone(self) -> typing.Any: ...
    @staticmethod
    def CreateSpecificCulture(name: str) -> CultureInfo: ...
    def Equals(self, value: typing.Any) -> bool: ...
    def GetConsoleFallbackUICulture(self) -> CultureInfo: ...
    @staticmethod
    def GetCultureInfoByIetfLanguageTag(name: str) -> CultureInfo: ...
    @staticmethod
    def GetCultures(types: CultureTypes) -> Array_1[CultureInfo]: ...
    def GetFormat(self, formatType: typing.Type[typing.Any]) -> typing.Any: ...
    def GetHashCode(self) -> int: ...
    @staticmethod
    def ReadOnly(ci: CultureInfo) -> CultureInfo: ...
    def ToString(self) -> str: ...
    # Skipped GetCultureInfo due to it being static, abstract and generic.

    GetCultureInfo : GetCultureInfo_MethodGroup
    class GetCultureInfo_MethodGroup:
        @typing.overload
        def __call__(self, culture: int) -> CultureInfo:...
        @typing.overload
        def __call__(self, name: str) -> CultureInfo:...
        @typing.overload
        def __call__(self, name: str, altName: str) -> CultureInfo:...
        @typing.overload
        def __call__(self, name: str, predefinedOnly: bool) -> CultureInfo:...



class CultureNotFoundException(ArgumentException):
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, message: str) -> None: ...
    @typing.overload
    def __init__(self, message: str, innerException: Exception) -> None: ...
    @typing.overload
    def __init__(self, message: str, invalidCultureId: int, innerException: Exception) -> None: ...
    @typing.overload
    def __init__(self, message: str, invalidCultureName: str, innerException: Exception) -> None: ...
    @typing.overload
    def __init__(self, paramName: str, invalidCultureId: int, message: str) -> None: ...
    @typing.overload
    def __init__(self, paramName: str, invalidCultureName: str, message: str) -> None: ...
    @typing.overload
    def __init__(self, paramName: str, message: str) -> None: ...
    @property
    def Data(self) -> IDictionary: ...
    @property
    def HelpLink(self) -> str: ...
    @HelpLink.setter
    def HelpLink(self, value: str) -> str: ...
    @property
    def HResult(self) -> int: ...
    @HResult.setter
    def HResult(self, value: int) -> int: ...
    @property
    def InnerException(self) -> Exception: ...
    @property
    def InvalidCultureId(self) -> typing.Optional[int]: ...
    @property
    def InvalidCultureName(self) -> str: ...
    @property
    def Message(self) -> str: ...
    @property
    def ParamName(self) -> str: ...
    @property
    def Source(self) -> str: ...
    @Source.setter
    def Source(self, value: str) -> str: ...
    @property
    def StackTrace(self) -> str: ...
    @property
    def TargetSite(self) -> MethodBase: ...
    def GetObjectData(self, info: SerializationInfo, context: StreamingContext) -> None: ...


class CultureTypes(typing.SupportsInt):
    @typing.overload
    def __init__(self, value : int) -> None: ...
    @typing.overload
    def __init__(self, value : int, force_if_true: bool) -> None: ...
    def __int__(self) -> int: ...
    
    # Values:
    NeutralCultures : CultureTypes # 1
    SpecificCultures : CultureTypes # 2
    InstalledWin32Cultures : CultureTypes # 4
    AllCultures : CultureTypes # 7
    UserCustomCulture : CultureTypes # 8
    ReplacementCultures : CultureTypes # 16
    WindowsOnlyCultures : CultureTypes # 32
    FrameworkCultures : CultureTypes # 64


class DateTimeFormatInfo(ICloneable, IFormatProvider):
    def __init__(self) -> None: ...
    @property
    def AbbreviatedDayNames(self) -> Array_1[str]: ...
    @AbbreviatedDayNames.setter
    def AbbreviatedDayNames(self, value: Array_1[str]) -> Array_1[str]: ...
    @property
    def AbbreviatedMonthGenitiveNames(self) -> Array_1[str]: ...
    @AbbreviatedMonthGenitiveNames.setter
    def AbbreviatedMonthGenitiveNames(self, value: Array_1[str]) -> Array_1[str]: ...
    @property
    def AbbreviatedMonthNames(self) -> Array_1[str]: ...
    @AbbreviatedMonthNames.setter
    def AbbreviatedMonthNames(self, value: Array_1[str]) -> Array_1[str]: ...
    @property
    def AMDesignator(self) -> str: ...
    @AMDesignator.setter
    def AMDesignator(self, value: str) -> str: ...
    @property
    def Calendar(self) -> Calendar: ...
    @Calendar.setter
    def Calendar(self, value: Calendar) -> Calendar: ...
    @property
    def CalendarWeekRule(self) -> CalendarWeekRule: ...
    @CalendarWeekRule.setter
    def CalendarWeekRule(self, value: CalendarWeekRule) -> CalendarWeekRule: ...
    @classmethod
    @property
    def CurrentInfo(cls) -> DateTimeFormatInfo: ...
    @property
    def DateSeparator(self) -> str: ...
    @DateSeparator.setter
    def DateSeparator(self, value: str) -> str: ...
    @property
    def DayNames(self) -> Array_1[str]: ...
    @DayNames.setter
    def DayNames(self, value: Array_1[str]) -> Array_1[str]: ...
    @property
    def FirstDayOfWeek(self) -> DayOfWeek: ...
    @FirstDayOfWeek.setter
    def FirstDayOfWeek(self, value: DayOfWeek) -> DayOfWeek: ...
    @property
    def FullDateTimePattern(self) -> str: ...
    @FullDateTimePattern.setter
    def FullDateTimePattern(self, value: str) -> str: ...
    @classmethod
    @property
    def InvariantInfo(cls) -> DateTimeFormatInfo: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def LongDatePattern(self) -> str: ...
    @LongDatePattern.setter
    def LongDatePattern(self, value: str) -> str: ...
    @property
    def LongTimePattern(self) -> str: ...
    @LongTimePattern.setter
    def LongTimePattern(self, value: str) -> str: ...
    @property
    def MonthDayPattern(self) -> str: ...
    @MonthDayPattern.setter
    def MonthDayPattern(self, value: str) -> str: ...
    @property
    def MonthGenitiveNames(self) -> Array_1[str]: ...
    @MonthGenitiveNames.setter
    def MonthGenitiveNames(self, value: Array_1[str]) -> Array_1[str]: ...
    @property
    def MonthNames(self) -> Array_1[str]: ...
    @MonthNames.setter
    def MonthNames(self, value: Array_1[str]) -> Array_1[str]: ...
    @property
    def NativeCalendarName(self) -> str: ...
    @property
    def PMDesignator(self) -> str: ...
    @PMDesignator.setter
    def PMDesignator(self, value: str) -> str: ...
    @property
    def RFC1123Pattern(self) -> str: ...
    @property
    def ShortDatePattern(self) -> str: ...
    @ShortDatePattern.setter
    def ShortDatePattern(self, value: str) -> str: ...
    @property
    def ShortestDayNames(self) -> Array_1[str]: ...
    @ShortestDayNames.setter
    def ShortestDayNames(self, value: Array_1[str]) -> Array_1[str]: ...
    @property
    def ShortTimePattern(self) -> str: ...
    @ShortTimePattern.setter
    def ShortTimePattern(self, value: str) -> str: ...
    @property
    def SortableDateTimePattern(self) -> str: ...
    @property
    def TimeSeparator(self) -> str: ...
    @TimeSeparator.setter
    def TimeSeparator(self, value: str) -> str: ...
    @property
    def UniversalSortableDateTimePattern(self) -> str: ...
    @property
    def YearMonthPattern(self) -> str: ...
    @YearMonthPattern.setter
    def YearMonthPattern(self, value: str) -> str: ...
    def Clone(self) -> typing.Any: ...
    def GetAbbreviatedDayName(self, dayofweek: DayOfWeek) -> str: ...
    def GetAbbreviatedEraName(self, era: int) -> str: ...
    def GetAbbreviatedMonthName(self, month: int) -> str: ...
    def GetDayName(self, dayofweek: DayOfWeek) -> str: ...
    def GetEra(self, eraName: str) -> int: ...
    def GetEraName(self, era: int) -> str: ...
    def GetFormat(self, formatType: typing.Type[typing.Any]) -> typing.Any: ...
    @staticmethod
    def GetInstance(provider: IFormatProvider) -> DateTimeFormatInfo: ...
    def GetMonthName(self, month: int) -> str: ...
    def GetShortestDayName(self, dayOfWeek: DayOfWeek) -> str: ...
    @staticmethod
    def ReadOnly(dtfi: DateTimeFormatInfo) -> DateTimeFormatInfo: ...
    def SetAllDateTimePatterns(self, patterns: Array_1[str], format: str) -> None: ...
    # Skipped GetAllDateTimePatterns due to it being static, abstract and generic.

    GetAllDateTimePatterns : GetAllDateTimePatterns_MethodGroup
    class GetAllDateTimePatterns_MethodGroup:
        @typing.overload
        def __call__(self) -> Array_1[str]:...
        @typing.overload
        def __call__(self, format: str) -> Array_1[str]:...



class DateTimeStyles(typing.SupportsInt):
    @typing.overload
    def __init__(self, value : int) -> None: ...
    @typing.overload
    def __init__(self, value : int, force_if_true: bool) -> None: ...
    def __int__(self) -> int: ...
    
    # Values:
    None_ : DateTimeStyles # 0
    AllowLeadingWhite : DateTimeStyles # 1
    AllowTrailingWhite : DateTimeStyles # 2
    AllowInnerWhite : DateTimeStyles # 4
    AllowWhiteSpaces : DateTimeStyles # 7
    NoCurrentDateDefault : DateTimeStyles # 8
    AdjustToUniversal : DateTimeStyles # 16
    AssumeLocal : DateTimeStyles # 32
    AssumeUniversal : DateTimeStyles # 64
    RoundtripKind : DateTimeStyles # 128


class DaylightTime:
    def __init__(self, start: DateTime, end: DateTime, delta: TimeSpan) -> None: ...
    @property
    def Delta(self) -> TimeSpan: ...
    @property
    def End(self) -> DateTime: ...
    @property
    def Start(self) -> DateTime: ...


class DigitShapes(typing.SupportsInt):
    @typing.overload
    def __init__(self, value : int) -> None: ...
    @typing.overload
    def __init__(self, value : int, force_if_true: bool) -> None: ...
    def __int__(self) -> int: ...
    
    # Values:
    Context : DigitShapes # 0
    None_ : DigitShapes # 1
    NativeNational : DigitShapes # 2


class EastAsianLunisolarCalendar(Calendar):
    @property
    def AlgorithmType(self) -> CalendarAlgorithmType: ...
    @property
    def Eras(self) -> Array_1[int]: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def MaxSupportedDateTime(self) -> DateTime: ...
    @property
    def MinSupportedDateTime(self) -> DateTime: ...
    @property
    def TwoDigitYearMax(self) -> int: ...
    @TwoDigitYearMax.setter
    def TwoDigitYearMax(self, value: int) -> int: ...
    def AddMonths(self, time: DateTime, months: int) -> DateTime: ...
    def AddYears(self, time: DateTime, years: int) -> DateTime: ...
    def GetCelestialStem(self, sexagenaryYear: int) -> int: ...
    def GetDayOfMonth(self, time: DateTime) -> int: ...
    def GetDayOfWeek(self, time: DateTime) -> DayOfWeek: ...
    def GetDayOfYear(self, time: DateTime) -> int: ...
    def GetDaysInMonth(self, year: int, month: int, era: int) -> int: ...
    def GetDaysInYear(self, year: int, era: int) -> int: ...
    def GetLeapMonth(self, year: int, era: int) -> int: ...
    def GetMonth(self, time: DateTime) -> int: ...
    def GetMonthsInYear(self, year: int, era: int) -> int: ...
    def GetSexagenaryYear(self, time: DateTime) -> int: ...
    def GetTerrestrialBranch(self, sexagenaryYear: int) -> int: ...
    def GetYear(self, time: DateTime) -> int: ...
    def IsLeapDay(self, year: int, month: int, day: int, era: int) -> bool: ...
    def IsLeapMonth(self, year: int, month: int, era: int) -> bool: ...
    def IsLeapYear(self, year: int, era: int) -> bool: ...
    def ToDateTime(self, year: int, month: int, day: int, hour: int, minute: int, second: int, millisecond: int, era: int) -> DateTime: ...
    def ToFourDigitYear(self, year: int) -> int: ...


class GlobalizationExtensions(abc.ABC):
    @staticmethod
    def GetStringComparer(compareInfo: CompareInfo, options: CompareOptions) -> StringComparer: ...


class GregorianCalendar(Calendar):
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, type: GregorianCalendarTypes) -> None: ...
    ADEra : int
    @property
    def AlgorithmType(self) -> CalendarAlgorithmType: ...
    @property
    def CalendarType(self) -> GregorianCalendarTypes: ...
    @CalendarType.setter
    def CalendarType(self, value: GregorianCalendarTypes) -> GregorianCalendarTypes: ...
    @property
    def Eras(self) -> Array_1[int]: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def MaxSupportedDateTime(self) -> DateTime: ...
    @property
    def MinSupportedDateTime(self) -> DateTime: ...
    @property
    def TwoDigitYearMax(self) -> int: ...
    @TwoDigitYearMax.setter
    def TwoDigitYearMax(self, value: int) -> int: ...
    def AddMonths(self, time: DateTime, months: int) -> DateTime: ...
    def AddYears(self, time: DateTime, years: int) -> DateTime: ...
    def GetDayOfMonth(self, time: DateTime) -> int: ...
    def GetDayOfWeek(self, time: DateTime) -> DayOfWeek: ...
    def GetDayOfYear(self, time: DateTime) -> int: ...
    def GetDaysInMonth(self, year: int, month: int, era: int) -> int: ...
    def GetDaysInYear(self, year: int, era: int) -> int: ...
    def GetEra(self, time: DateTime) -> int: ...
    def GetLeapMonth(self, year: int, era: int) -> int: ...
    def GetMonth(self, time: DateTime) -> int: ...
    def GetMonthsInYear(self, year: int, era: int) -> int: ...
    def GetYear(self, time: DateTime) -> int: ...
    def IsLeapDay(self, year: int, month: int, day: int, era: int) -> bool: ...
    def IsLeapMonth(self, year: int, month: int, era: int) -> bool: ...
    def IsLeapYear(self, year: int, era: int) -> bool: ...
    def ToDateTime(self, year: int, month: int, day: int, hour: int, minute: int, second: int, millisecond: int, era: int) -> DateTime: ...
    def ToFourDigitYear(self, year: int) -> int: ...


class GregorianCalendarTypes(typing.SupportsInt):
    @typing.overload
    def __init__(self, value : int) -> None: ...
    @typing.overload
    def __init__(self, value : int, force_if_true: bool) -> None: ...
    def __int__(self) -> int: ...
    
    # Values:
    Localized : GregorianCalendarTypes # 1
    USEnglish : GregorianCalendarTypes # 2
    MiddleEastFrench : GregorianCalendarTypes # 9
    Arabic : GregorianCalendarTypes # 10
    TransliteratedEnglish : GregorianCalendarTypes # 11
    TransliteratedFrench : GregorianCalendarTypes # 12


class HebrewCalendar(Calendar):
    def __init__(self) -> None: ...
    HebrewEra : int
    @property
    def AlgorithmType(self) -> CalendarAlgorithmType: ...
    @property
    def Eras(self) -> Array_1[int]: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def MaxSupportedDateTime(self) -> DateTime: ...
    @property
    def MinSupportedDateTime(self) -> DateTime: ...
    @property
    def TwoDigitYearMax(self) -> int: ...
    @TwoDigitYearMax.setter
    def TwoDigitYearMax(self, value: int) -> int: ...
    def AddMonths(self, time: DateTime, months: int) -> DateTime: ...
    def AddYears(self, time: DateTime, years: int) -> DateTime: ...
    def GetDayOfMonth(self, time: DateTime) -> int: ...
    def GetDayOfWeek(self, time: DateTime) -> DayOfWeek: ...
    def GetDayOfYear(self, time: DateTime) -> int: ...
    def GetDaysInMonth(self, year: int, month: int, era: int) -> int: ...
    def GetDaysInYear(self, year: int, era: int) -> int: ...
    def GetEra(self, time: DateTime) -> int: ...
    def GetLeapMonth(self, year: int, era: int) -> int: ...
    def GetMonth(self, time: DateTime) -> int: ...
    def GetMonthsInYear(self, year: int, era: int) -> int: ...
    def GetYear(self, time: DateTime) -> int: ...
    def IsLeapDay(self, year: int, month: int, day: int, era: int) -> bool: ...
    def IsLeapMonth(self, year: int, month: int, era: int) -> bool: ...
    def IsLeapYear(self, year: int, era: int) -> bool: ...
    def ToDateTime(self, year: int, month: int, day: int, hour: int, minute: int, second: int, millisecond: int, era: int) -> DateTime: ...
    def ToFourDigitYear(self, year: int) -> int: ...


class HijriCalendar(Calendar):
    def __init__(self) -> None: ...
    HijriEra : int
    @property
    def AlgorithmType(self) -> CalendarAlgorithmType: ...
    @property
    def Eras(self) -> Array_1[int]: ...
    @property
    def HijriAdjustment(self) -> int: ...
    @HijriAdjustment.setter
    def HijriAdjustment(self, value: int) -> int: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def MaxSupportedDateTime(self) -> DateTime: ...
    @property
    def MinSupportedDateTime(self) -> DateTime: ...
    @property
    def TwoDigitYearMax(self) -> int: ...
    @TwoDigitYearMax.setter
    def TwoDigitYearMax(self, value: int) -> int: ...
    def AddMonths(self, time: DateTime, months: int) -> DateTime: ...
    def AddYears(self, time: DateTime, years: int) -> DateTime: ...
    def GetDayOfMonth(self, time: DateTime) -> int: ...
    def GetDayOfWeek(self, time: DateTime) -> DayOfWeek: ...
    def GetDayOfYear(self, time: DateTime) -> int: ...
    def GetDaysInMonth(self, year: int, month: int, era: int) -> int: ...
    def GetDaysInYear(self, year: int, era: int) -> int: ...
    def GetEra(self, time: DateTime) -> int: ...
    def GetLeapMonth(self, year: int, era: int) -> int: ...
    def GetMonth(self, time: DateTime) -> int: ...
    def GetMonthsInYear(self, year: int, era: int) -> int: ...
    def GetYear(self, time: DateTime) -> int: ...
    def IsLeapDay(self, year: int, month: int, day: int, era: int) -> bool: ...
    def IsLeapMonth(self, year: int, month: int, era: int) -> bool: ...
    def IsLeapYear(self, year: int, era: int) -> bool: ...
    def ToDateTime(self, year: int, month: int, day: int, hour: int, minute: int, second: int, millisecond: int, era: int) -> DateTime: ...
    def ToFourDigitYear(self, year: int) -> int: ...


class IdnMapping:
    def __init__(self) -> None: ...
    @property
    def AllowUnassigned(self) -> bool: ...
    @AllowUnassigned.setter
    def AllowUnassigned(self, value: bool) -> bool: ...
    @property
    def UseStd3AsciiRules(self) -> bool: ...
    @UseStd3AsciiRules.setter
    def UseStd3AsciiRules(self, value: bool) -> bool: ...
    def Equals(self, obj: typing.Any) -> bool: ...
    def GetHashCode(self) -> int: ...
    # Skipped GetAscii due to it being static, abstract and generic.

    GetAscii : GetAscii_MethodGroup
    class GetAscii_MethodGroup:
        @typing.overload
        def __call__(self, unicode: str) -> str:...
        @typing.overload
        def __call__(self, unicode: str, index: int) -> str:...
        @typing.overload
        def __call__(self, unicode: str, index: int, count: int) -> str:...

    # Skipped GetUnicode due to it being static, abstract and generic.

    GetUnicode : GetUnicode_MethodGroup
    class GetUnicode_MethodGroup:
        @typing.overload
        def __call__(self, ascii: str) -> str:...
        @typing.overload
        def __call__(self, ascii: str, index: int) -> str:...
        @typing.overload
        def __call__(self, ascii: str, index: int, count: int) -> str:...



class ISOWeek(abc.ABC):
    @staticmethod
    def GetWeekOfYear(date: DateTime) -> int: ...
    @staticmethod
    def GetWeeksInYear(year: int) -> int: ...
    @staticmethod
    def GetYear(date: DateTime) -> int: ...
    @staticmethod
    def GetYearEnd(year: int) -> DateTime: ...
    @staticmethod
    def GetYearStart(year: int) -> DateTime: ...
    @staticmethod
    def ToDateTime(year: int, week: int, dayOfWeek: DayOfWeek) -> DateTime: ...


class JapaneseCalendar(Calendar):
    def __init__(self) -> None: ...
    @property
    def AlgorithmType(self) -> CalendarAlgorithmType: ...
    @property
    def Eras(self) -> Array_1[int]: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def MaxSupportedDateTime(self) -> DateTime: ...
    @property
    def MinSupportedDateTime(self) -> DateTime: ...
    @property
    def TwoDigitYearMax(self) -> int: ...
    @TwoDigitYearMax.setter
    def TwoDigitYearMax(self, value: int) -> int: ...
    def AddMonths(self, time: DateTime, months: int) -> DateTime: ...
    def AddYears(self, time: DateTime, years: int) -> DateTime: ...
    def GetDayOfMonth(self, time: DateTime) -> int: ...
    def GetDayOfWeek(self, time: DateTime) -> DayOfWeek: ...
    def GetDayOfYear(self, time: DateTime) -> int: ...
    def GetDaysInMonth(self, year: int, month: int, era: int) -> int: ...
    def GetDaysInYear(self, year: int, era: int) -> int: ...
    def GetEra(self, time: DateTime) -> int: ...
    def GetLeapMonth(self, year: int, era: int) -> int: ...
    def GetMonth(self, time: DateTime) -> int: ...
    def GetMonthsInYear(self, year: int, era: int) -> int: ...
    def GetWeekOfYear(self, time: DateTime, rule: CalendarWeekRule, firstDayOfWeek: DayOfWeek) -> int: ...
    def GetYear(self, time: DateTime) -> int: ...
    def IsLeapDay(self, year: int, month: int, day: int, era: int) -> bool: ...
    def IsLeapMonth(self, year: int, month: int, era: int) -> bool: ...
    def IsLeapYear(self, year: int, era: int) -> bool: ...
    def ToDateTime(self, year: int, month: int, day: int, hour: int, minute: int, second: int, millisecond: int, era: int) -> DateTime: ...
    def ToFourDigitYear(self, year: int) -> int: ...


class JapaneseLunisolarCalendar(EastAsianLunisolarCalendar):
    def __init__(self) -> None: ...
    JapaneseEra : int
    @property
    def AlgorithmType(self) -> CalendarAlgorithmType: ...
    @property
    def Eras(self) -> Array_1[int]: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def MaxSupportedDateTime(self) -> DateTime: ...
    @property
    def MinSupportedDateTime(self) -> DateTime: ...
    @property
    def TwoDigitYearMax(self) -> int: ...
    @TwoDigitYearMax.setter
    def TwoDigitYearMax(self, value: int) -> int: ...
    def GetEra(self, time: DateTime) -> int: ...


class JulianCalendar(Calendar):
    def __init__(self) -> None: ...
    JulianEra : int
    @property
    def AlgorithmType(self) -> CalendarAlgorithmType: ...
    @property
    def Eras(self) -> Array_1[int]: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def MaxSupportedDateTime(self) -> DateTime: ...
    @property
    def MinSupportedDateTime(self) -> DateTime: ...
    @property
    def TwoDigitYearMax(self) -> int: ...
    @TwoDigitYearMax.setter
    def TwoDigitYearMax(self, value: int) -> int: ...
    def AddMonths(self, time: DateTime, months: int) -> DateTime: ...
    def AddYears(self, time: DateTime, years: int) -> DateTime: ...
    def GetDayOfMonth(self, time: DateTime) -> int: ...
    def GetDayOfWeek(self, time: DateTime) -> DayOfWeek: ...
    def GetDayOfYear(self, time: DateTime) -> int: ...
    def GetDaysInMonth(self, year: int, month: int, era: int) -> int: ...
    def GetDaysInYear(self, year: int, era: int) -> int: ...
    def GetEra(self, time: DateTime) -> int: ...
    def GetLeapMonth(self, year: int, era: int) -> int: ...
    def GetMonth(self, time: DateTime) -> int: ...
    def GetMonthsInYear(self, year: int, era: int) -> int: ...
    def GetYear(self, time: DateTime) -> int: ...
    def IsLeapDay(self, year: int, month: int, day: int, era: int) -> bool: ...
    def IsLeapMonth(self, year: int, month: int, era: int) -> bool: ...
    def IsLeapYear(self, year: int, era: int) -> bool: ...
    def ToDateTime(self, year: int, month: int, day: int, hour: int, minute: int, second: int, millisecond: int, era: int) -> DateTime: ...
    def ToFourDigitYear(self, year: int) -> int: ...


class KoreanCalendar(Calendar):
    def __init__(self) -> None: ...
    KoreanEra : int
    @property
    def AlgorithmType(self) -> CalendarAlgorithmType: ...
    @property
    def Eras(self) -> Array_1[int]: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def MaxSupportedDateTime(self) -> DateTime: ...
    @property
    def MinSupportedDateTime(self) -> DateTime: ...
    @property
    def TwoDigitYearMax(self) -> int: ...
    @TwoDigitYearMax.setter
    def TwoDigitYearMax(self, value: int) -> int: ...
    def AddMonths(self, time: DateTime, months: int) -> DateTime: ...
    def AddYears(self, time: DateTime, years: int) -> DateTime: ...
    def GetDayOfMonth(self, time: DateTime) -> int: ...
    def GetDayOfWeek(self, time: DateTime) -> DayOfWeek: ...
    def GetDayOfYear(self, time: DateTime) -> int: ...
    def GetDaysInMonth(self, year: int, month: int, era: int) -> int: ...
    def GetDaysInYear(self, year: int, era: int) -> int: ...
    def GetEra(self, time: DateTime) -> int: ...
    def GetLeapMonth(self, year: int, era: int) -> int: ...
    def GetMonth(self, time: DateTime) -> int: ...
    def GetMonthsInYear(self, year: int, era: int) -> int: ...
    def GetWeekOfYear(self, time: DateTime, rule: CalendarWeekRule, firstDayOfWeek: DayOfWeek) -> int: ...
    def GetYear(self, time: DateTime) -> int: ...
    def IsLeapDay(self, year: int, month: int, day: int, era: int) -> bool: ...
    def IsLeapMonth(self, year: int, month: int, era: int) -> bool: ...
    def IsLeapYear(self, year: int, era: int) -> bool: ...
    def ToDateTime(self, year: int, month: int, day: int, hour: int, minute: int, second: int, millisecond: int, era: int) -> DateTime: ...
    def ToFourDigitYear(self, year: int) -> int: ...


class KoreanLunisolarCalendar(EastAsianLunisolarCalendar):
    def __init__(self) -> None: ...
    GregorianEra : int
    @property
    def AlgorithmType(self) -> CalendarAlgorithmType: ...
    @property
    def Eras(self) -> Array_1[int]: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def MaxSupportedDateTime(self) -> DateTime: ...
    @property
    def MinSupportedDateTime(self) -> DateTime: ...
    @property
    def TwoDigitYearMax(self) -> int: ...
    @TwoDigitYearMax.setter
    def TwoDigitYearMax(self, value: int) -> int: ...
    def GetEra(self, time: DateTime) -> int: ...


class NumberFormatInfo(ICloneable, IFormatProvider):
    def __init__(self) -> None: ...
    @property
    def CurrencyDecimalDigits(self) -> int: ...
    @CurrencyDecimalDigits.setter
    def CurrencyDecimalDigits(self, value: int) -> int: ...
    @property
    def CurrencyDecimalSeparator(self) -> str: ...
    @CurrencyDecimalSeparator.setter
    def CurrencyDecimalSeparator(self, value: str) -> str: ...
    @property
    def CurrencyGroupSeparator(self) -> str: ...
    @CurrencyGroupSeparator.setter
    def CurrencyGroupSeparator(self, value: str) -> str: ...
    @property
    def CurrencyGroupSizes(self) -> Array_1[int]: ...
    @CurrencyGroupSizes.setter
    def CurrencyGroupSizes(self, value: Array_1[int]) -> Array_1[int]: ...
    @property
    def CurrencyNegativePattern(self) -> int: ...
    @CurrencyNegativePattern.setter
    def CurrencyNegativePattern(self, value: int) -> int: ...
    @property
    def CurrencyPositivePattern(self) -> int: ...
    @CurrencyPositivePattern.setter
    def CurrencyPositivePattern(self, value: int) -> int: ...
    @property
    def CurrencySymbol(self) -> str: ...
    @CurrencySymbol.setter
    def CurrencySymbol(self, value: str) -> str: ...
    @classmethod
    @property
    def CurrentInfo(cls) -> NumberFormatInfo: ...
    @property
    def DigitSubstitution(self) -> DigitShapes: ...
    @DigitSubstitution.setter
    def DigitSubstitution(self, value: DigitShapes) -> DigitShapes: ...
    @classmethod
    @property
    def InvariantInfo(cls) -> NumberFormatInfo: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def NaNSymbol(self) -> str: ...
    @NaNSymbol.setter
    def NaNSymbol(self, value: str) -> str: ...
    @property
    def NativeDigits(self) -> Array_1[str]: ...
    @NativeDigits.setter
    def NativeDigits(self, value: Array_1[str]) -> Array_1[str]: ...
    @property
    def NegativeInfinitySymbol(self) -> str: ...
    @NegativeInfinitySymbol.setter
    def NegativeInfinitySymbol(self, value: str) -> str: ...
    @property
    def NegativeSign(self) -> str: ...
    @NegativeSign.setter
    def NegativeSign(self, value: str) -> str: ...
    @property
    def NumberDecimalDigits(self) -> int: ...
    @NumberDecimalDigits.setter
    def NumberDecimalDigits(self, value: int) -> int: ...
    @property
    def NumberDecimalSeparator(self) -> str: ...
    @NumberDecimalSeparator.setter
    def NumberDecimalSeparator(self, value: str) -> str: ...
    @property
    def NumberGroupSeparator(self) -> str: ...
    @NumberGroupSeparator.setter
    def NumberGroupSeparator(self, value: str) -> str: ...
    @property
    def NumberGroupSizes(self) -> Array_1[int]: ...
    @NumberGroupSizes.setter
    def NumberGroupSizes(self, value: Array_1[int]) -> Array_1[int]: ...
    @property
    def NumberNegativePattern(self) -> int: ...
    @NumberNegativePattern.setter
    def NumberNegativePattern(self, value: int) -> int: ...
    @property
    def PercentDecimalDigits(self) -> int: ...
    @PercentDecimalDigits.setter
    def PercentDecimalDigits(self, value: int) -> int: ...
    @property
    def PercentDecimalSeparator(self) -> str: ...
    @PercentDecimalSeparator.setter
    def PercentDecimalSeparator(self, value: str) -> str: ...
    @property
    def PercentGroupSeparator(self) -> str: ...
    @PercentGroupSeparator.setter
    def PercentGroupSeparator(self, value: str) -> str: ...
    @property
    def PercentGroupSizes(self) -> Array_1[int]: ...
    @PercentGroupSizes.setter
    def PercentGroupSizes(self, value: Array_1[int]) -> Array_1[int]: ...
    @property
    def PercentNegativePattern(self) -> int: ...
    @PercentNegativePattern.setter
    def PercentNegativePattern(self, value: int) -> int: ...
    @property
    def PercentPositivePattern(self) -> int: ...
    @PercentPositivePattern.setter
    def PercentPositivePattern(self, value: int) -> int: ...
    @property
    def PercentSymbol(self) -> str: ...
    @PercentSymbol.setter
    def PercentSymbol(self, value: str) -> str: ...
    @property
    def PerMilleSymbol(self) -> str: ...
    @PerMilleSymbol.setter
    def PerMilleSymbol(self, value: str) -> str: ...
    @property
    def PositiveInfinitySymbol(self) -> str: ...
    @PositiveInfinitySymbol.setter
    def PositiveInfinitySymbol(self, value: str) -> str: ...
    @property
    def PositiveSign(self) -> str: ...
    @PositiveSign.setter
    def PositiveSign(self, value: str) -> str: ...
    def Clone(self) -> typing.Any: ...
    def GetFormat(self, formatType: typing.Type[typing.Any]) -> typing.Any: ...
    @staticmethod
    def GetInstance(formatProvider: IFormatProvider) -> NumberFormatInfo: ...
    @staticmethod
    def ReadOnly(nfi: NumberFormatInfo) -> NumberFormatInfo: ...


class NumberStyles(typing.SupportsInt):
    @typing.overload
    def __init__(self, value : int) -> None: ...
    @typing.overload
    def __init__(self, value : int, force_if_true: bool) -> None: ...
    def __int__(self) -> int: ...
    
    # Values:
    None_ : NumberStyles # 0
    AllowLeadingWhite : NumberStyles # 1
    AllowTrailingWhite : NumberStyles # 2
    AllowLeadingSign : NumberStyles # 4
    Integer : NumberStyles # 7
    AllowTrailingSign : NumberStyles # 8
    AllowParentheses : NumberStyles # 16
    AllowDecimalPoint : NumberStyles # 32
    AllowThousands : NumberStyles # 64
    Number : NumberStyles # 111
    AllowExponent : NumberStyles # 128
    Float : NumberStyles # 167
    AllowCurrencySymbol : NumberStyles # 256
    Currency : NumberStyles # 383
    Any : NumberStyles # 511
    AllowHexSpecifier : NumberStyles # 512
    HexNumber : NumberStyles # 515
    AllowBinarySpecifier : NumberStyles # 1024
    BinaryNumber : NumberStyles # 1027


class PersianCalendar(Calendar):
    def __init__(self) -> None: ...
    PersianEra : int
    @property
    def AlgorithmType(self) -> CalendarAlgorithmType: ...
    @property
    def Eras(self) -> Array_1[int]: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def MaxSupportedDateTime(self) -> DateTime: ...
    @property
    def MinSupportedDateTime(self) -> DateTime: ...
    @property
    def TwoDigitYearMax(self) -> int: ...
    @TwoDigitYearMax.setter
    def TwoDigitYearMax(self, value: int) -> int: ...
    def AddMonths(self, time: DateTime, months: int) -> DateTime: ...
    def AddYears(self, time: DateTime, years: int) -> DateTime: ...
    def GetDayOfMonth(self, time: DateTime) -> int: ...
    def GetDayOfWeek(self, time: DateTime) -> DayOfWeek: ...
    def GetDayOfYear(self, time: DateTime) -> int: ...
    def GetDaysInMonth(self, year: int, month: int, era: int) -> int: ...
    def GetDaysInYear(self, year: int, era: int) -> int: ...
    def GetEra(self, time: DateTime) -> int: ...
    def GetLeapMonth(self, year: int, era: int) -> int: ...
    def GetMonth(self, time: DateTime) -> int: ...
    def GetMonthsInYear(self, year: int, era: int) -> int: ...
    def GetYear(self, time: DateTime) -> int: ...
    def IsLeapDay(self, year: int, month: int, day: int, era: int) -> bool: ...
    def IsLeapMonth(self, year: int, month: int, era: int) -> bool: ...
    def IsLeapYear(self, year: int, era: int) -> bool: ...
    def ToDateTime(self, year: int, month: int, day: int, hour: int, minute: int, second: int, millisecond: int, era: int) -> DateTime: ...
    def ToFourDigitYear(self, year: int) -> int: ...


class RegionInfo:
    @typing.overload
    def __init__(self, culture: int) -> None: ...
    @typing.overload
    def __init__(self, name: str) -> None: ...
    @property
    def CurrencyEnglishName(self) -> str: ...
    @property
    def CurrencyNativeName(self) -> str: ...
    @property
    def CurrencySymbol(self) -> str: ...
    @classmethod
    @property
    def CurrentRegion(cls) -> RegionInfo: ...
    @property
    def DisplayName(self) -> str: ...
    @property
    def EnglishName(self) -> str: ...
    @property
    def GeoId(self) -> int: ...
    @property
    def IsMetric(self) -> bool: ...
    @property
    def ISOCurrencySymbol(self) -> str: ...
    @property
    def Name(self) -> str: ...
    @property
    def NativeName(self) -> str: ...
    @property
    def ThreeLetterISORegionName(self) -> str: ...
    @property
    def ThreeLetterWindowsRegionName(self) -> str: ...
    @property
    def TwoLetterISORegionName(self) -> str: ...
    def Equals(self, value: typing.Any) -> bool: ...
    def GetHashCode(self) -> int: ...
    def ToString(self) -> str: ...


class SortKey:
    @property
    def KeyData(self) -> Array_1[int]: ...
    @property
    def OriginalString(self) -> str: ...
    @staticmethod
    def Compare(sortkey1: SortKey, sortkey2: SortKey) -> int: ...
    def Equals(self, value: typing.Any) -> bool: ...
    def GetHashCode(self) -> int: ...
    def ToString(self) -> str: ...


class SortVersion(IEquatable_1[SortVersion]):
    def __init__(self, fullVersion: int, sortId: Guid) -> None: ...
    @property
    def FullVersion(self) -> int: ...
    @property
    def SortId(self) -> Guid: ...
    def GetHashCode(self) -> int: ...
    def __eq__(self, left: SortVersion, right: SortVersion) -> bool: ...
    def __ne__(self, left: SortVersion, right: SortVersion) -> bool: ...
    # Skipped Equals due to it being static, abstract and generic.

    Equals : Equals_MethodGroup
    class Equals_MethodGroup:
        @typing.overload
        def __call__(self, other: SortVersion) -> bool:...
        @typing.overload
        def __call__(self, obj: typing.Any) -> bool:...



class StringInfo:
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, value: str) -> None: ...
    @property
    def LengthInTextElements(self) -> int: ...
    @property
    def String(self) -> str: ...
    @String.setter
    def String(self, value: str) -> str: ...
    def Equals(self, value: typing.Any) -> bool: ...
    def GetHashCode(self) -> int: ...
    @staticmethod
    def ParseCombiningCharacters(str: str) -> Array_1[int]: ...
    # Skipped GetNextTextElement due to it being static, abstract and generic.

    GetNextTextElement : GetNextTextElement_MethodGroup
    class GetNextTextElement_MethodGroup:
        @typing.overload
        def __call__(self, str: str) -> str:...
        @typing.overload
        def __call__(self, str: str, index: int) -> str:...

    # Skipped GetNextTextElementLength due to it being static, abstract and generic.

    GetNextTextElementLength : GetNextTextElementLength_MethodGroup
    class GetNextTextElementLength_MethodGroup:
        @typing.overload
        def __call__(self, str: ReadOnlySpan_1[str]) -> int:...
        @typing.overload
        def __call__(self, str: str) -> int:...
        @typing.overload
        def __call__(self, str: str, index: int) -> int:...

    # Skipped GetTextElementEnumerator due to it being static, abstract and generic.

    GetTextElementEnumerator : GetTextElementEnumerator_MethodGroup
    class GetTextElementEnumerator_MethodGroup:
        @typing.overload
        def __call__(self, str: str) -> TextElementEnumerator:...
        @typing.overload
        def __call__(self, str: str, index: int) -> TextElementEnumerator:...

    # Skipped SubstringByTextElements due to it being static, abstract and generic.

    SubstringByTextElements : SubstringByTextElements_MethodGroup
    class SubstringByTextElements_MethodGroup:
        @typing.overload
        def __call__(self, startingTextElement: int) -> str:...
        @typing.overload
        def __call__(self, startingTextElement: int, lengthInTextElements: int) -> str:...



class TaiwanCalendar(Calendar):
    def __init__(self) -> None: ...
    @property
    def AlgorithmType(self) -> CalendarAlgorithmType: ...
    @property
    def Eras(self) -> Array_1[int]: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def MaxSupportedDateTime(self) -> DateTime: ...
    @property
    def MinSupportedDateTime(self) -> DateTime: ...
    @property
    def TwoDigitYearMax(self) -> int: ...
    @TwoDigitYearMax.setter
    def TwoDigitYearMax(self, value: int) -> int: ...
    def AddMonths(self, time: DateTime, months: int) -> DateTime: ...
    def AddYears(self, time: DateTime, years: int) -> DateTime: ...
    def GetDayOfMonth(self, time: DateTime) -> int: ...
    def GetDayOfWeek(self, time: DateTime) -> DayOfWeek: ...
    def GetDayOfYear(self, time: DateTime) -> int: ...
    def GetDaysInMonth(self, year: int, month: int, era: int) -> int: ...
    def GetDaysInYear(self, year: int, era: int) -> int: ...
    def GetEra(self, time: DateTime) -> int: ...
    def GetLeapMonth(self, year: int, era: int) -> int: ...
    def GetMonth(self, time: DateTime) -> int: ...
    def GetMonthsInYear(self, year: int, era: int) -> int: ...
    def GetWeekOfYear(self, time: DateTime, rule: CalendarWeekRule, firstDayOfWeek: DayOfWeek) -> int: ...
    def GetYear(self, time: DateTime) -> int: ...
    def IsLeapDay(self, year: int, month: int, day: int, era: int) -> bool: ...
    def IsLeapMonth(self, year: int, month: int, era: int) -> bool: ...
    def IsLeapYear(self, year: int, era: int) -> bool: ...
    def ToDateTime(self, year: int, month: int, day: int, hour: int, minute: int, second: int, millisecond: int, era: int) -> DateTime: ...
    def ToFourDigitYear(self, year: int) -> int: ...


class TaiwanLunisolarCalendar(EastAsianLunisolarCalendar):
    def __init__(self) -> None: ...
    @property
    def AlgorithmType(self) -> CalendarAlgorithmType: ...
    @property
    def Eras(self) -> Array_1[int]: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def MaxSupportedDateTime(self) -> DateTime: ...
    @property
    def MinSupportedDateTime(self) -> DateTime: ...
    @property
    def TwoDigitYearMax(self) -> int: ...
    @TwoDigitYearMax.setter
    def TwoDigitYearMax(self, value: int) -> int: ...
    def GetEra(self, time: DateTime) -> int: ...


class TextElementEnumerator(IEnumerator):
    @property
    def Current(self) -> typing.Any: ...
    @property
    def ElementIndex(self) -> int: ...
    def GetTextElement(self) -> str: ...
    def MoveNext(self) -> bool: ...
    def Reset(self) -> None: ...


class TextInfo(IDeserializationCallback, ICloneable):
    @property
    def ANSICodePage(self) -> int: ...
    @property
    def CultureName(self) -> str: ...
    @property
    def EBCDICCodePage(self) -> int: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def IsRightToLeft(self) -> bool: ...
    @property
    def LCID(self) -> int: ...
    @property
    def ListSeparator(self) -> str: ...
    @ListSeparator.setter
    def ListSeparator(self, value: str) -> str: ...
    @property
    def MacCodePage(self) -> int: ...
    @property
    def OEMCodePage(self) -> int: ...
    def Clone(self) -> typing.Any: ...
    def Equals(self, obj: typing.Any) -> bool: ...
    def GetHashCode(self) -> int: ...
    @staticmethod
    def ReadOnly(textInfo: TextInfo) -> TextInfo: ...
    def ToString(self) -> str: ...
    def ToTitleCase(self, str: str) -> str: ...
    # Skipped ToLower due to it being static, abstract and generic.

    ToLower : ToLower_MethodGroup
    class ToLower_MethodGroup:
        def __call__(self, c: str) -> str:...
        # Method ToLower(str : String) was skipped since it collides with above method

    # Skipped ToUpper due to it being static, abstract and generic.

    ToUpper : ToUpper_MethodGroup
    class ToUpper_MethodGroup:
        def __call__(self, c: str) -> str:...
        # Method ToUpper(str : String) was skipped since it collides with above method



class ThaiBuddhistCalendar(Calendar):
    def __init__(self) -> None: ...
    ThaiBuddhistEra : int
    @property
    def AlgorithmType(self) -> CalendarAlgorithmType: ...
    @property
    def Eras(self) -> Array_1[int]: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def MaxSupportedDateTime(self) -> DateTime: ...
    @property
    def MinSupportedDateTime(self) -> DateTime: ...
    @property
    def TwoDigitYearMax(self) -> int: ...
    @TwoDigitYearMax.setter
    def TwoDigitYearMax(self, value: int) -> int: ...
    def AddMonths(self, time: DateTime, months: int) -> DateTime: ...
    def AddYears(self, time: DateTime, years: int) -> DateTime: ...
    def GetDayOfMonth(self, time: DateTime) -> int: ...
    def GetDayOfWeek(self, time: DateTime) -> DayOfWeek: ...
    def GetDayOfYear(self, time: DateTime) -> int: ...
    def GetDaysInMonth(self, year: int, month: int, era: int) -> int: ...
    def GetDaysInYear(self, year: int, era: int) -> int: ...
    def GetEra(self, time: DateTime) -> int: ...
    def GetLeapMonth(self, year: int, era: int) -> int: ...
    def GetMonth(self, time: DateTime) -> int: ...
    def GetMonthsInYear(self, year: int, era: int) -> int: ...
    def GetWeekOfYear(self, time: DateTime, rule: CalendarWeekRule, firstDayOfWeek: DayOfWeek) -> int: ...
    def GetYear(self, time: DateTime) -> int: ...
    def IsLeapDay(self, year: int, month: int, day: int, era: int) -> bool: ...
    def IsLeapMonth(self, year: int, month: int, era: int) -> bool: ...
    def IsLeapYear(self, year: int, era: int) -> bool: ...
    def ToDateTime(self, year: int, month: int, day: int, hour: int, minute: int, second: int, millisecond: int, era: int) -> DateTime: ...
    def ToFourDigitYear(self, year: int) -> int: ...


class TimeSpanStyles(typing.SupportsInt):
    @typing.overload
    def __init__(self, value : int) -> None: ...
    @typing.overload
    def __init__(self, value : int, force_if_true: bool) -> None: ...
    def __int__(self) -> int: ...
    
    # Values:
    None_ : TimeSpanStyles # 0
    AssumeNegative : TimeSpanStyles # 1


class UmAlQuraCalendar(Calendar):
    def __init__(self) -> None: ...
    UmAlQuraEra : int
    @property
    def AlgorithmType(self) -> CalendarAlgorithmType: ...
    @property
    def Eras(self) -> Array_1[int]: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def MaxSupportedDateTime(self) -> DateTime: ...
    @property
    def MinSupportedDateTime(self) -> DateTime: ...
    @property
    def TwoDigitYearMax(self) -> int: ...
    @TwoDigitYearMax.setter
    def TwoDigitYearMax(self, value: int) -> int: ...
    def AddMonths(self, time: DateTime, months: int) -> DateTime: ...
    def AddYears(self, time: DateTime, years: int) -> DateTime: ...
    def GetDayOfMonth(self, time: DateTime) -> int: ...
    def GetDayOfWeek(self, time: DateTime) -> DayOfWeek: ...
    def GetDayOfYear(self, time: DateTime) -> int: ...
    def GetDaysInMonth(self, year: int, month: int, era: int) -> int: ...
    def GetDaysInYear(self, year: int, era: int) -> int: ...
    def GetEra(self, time: DateTime) -> int: ...
    def GetLeapMonth(self, year: int, era: int) -> int: ...
    def GetMonth(self, time: DateTime) -> int: ...
    def GetMonthsInYear(self, year: int, era: int) -> int: ...
    def GetYear(self, time: DateTime) -> int: ...
    def IsLeapDay(self, year: int, month: int, day: int, era: int) -> bool: ...
    def IsLeapMonth(self, year: int, month: int, era: int) -> bool: ...
    def IsLeapYear(self, year: int, era: int) -> bool: ...
    def ToDateTime(self, year: int, month: int, day: int, hour: int, minute: int, second: int, millisecond: int, era: int) -> DateTime: ...
    def ToFourDigitYear(self, year: int) -> int: ...


class UnicodeCategory(typing.SupportsInt):
    @typing.overload
    def __init__(self, value : int) -> None: ...
    @typing.overload
    def __init__(self, value : int, force_if_true: bool) -> None: ...
    def __int__(self) -> int: ...
    
    # Values:
    UppercaseLetter : UnicodeCategory # 0
    LowercaseLetter : UnicodeCategory # 1
    TitlecaseLetter : UnicodeCategory # 2
    ModifierLetter : UnicodeCategory # 3
    OtherLetter : UnicodeCategory # 4
    NonSpacingMark : UnicodeCategory # 5
    SpacingCombiningMark : UnicodeCategory # 6
    EnclosingMark : UnicodeCategory # 7
    DecimalDigitNumber : UnicodeCategory # 8
    LetterNumber : UnicodeCategory # 9
    OtherNumber : UnicodeCategory # 10
    SpaceSeparator : UnicodeCategory # 11
    LineSeparator : UnicodeCategory # 12
    ParagraphSeparator : UnicodeCategory # 13
    Control : UnicodeCategory # 14
    Format : UnicodeCategory # 15
    Surrogate : UnicodeCategory # 16
    PrivateUse : UnicodeCategory # 17
    ConnectorPunctuation : UnicodeCategory # 18
    DashPunctuation : UnicodeCategory # 19
    OpenPunctuation : UnicodeCategory # 20
    ClosePunctuation : UnicodeCategory # 21
    InitialQuotePunctuation : UnicodeCategory # 22
    FinalQuotePunctuation : UnicodeCategory # 23
    OtherPunctuation : UnicodeCategory # 24
    MathSymbol : UnicodeCategory # 25
    CurrencySymbol : UnicodeCategory # 26
    ModifierSymbol : UnicodeCategory # 27
    OtherSymbol : UnicodeCategory # 28
    OtherNotAssigned : UnicodeCategory # 29




from src.PySimultan2 import DataModel
from src.PySimultan2.utils import add_sub_component

try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 importlib_resources.
    import importlib_resources as pkg_resources


import PySimultan2.tests.readme_examples.resources as readme_examples

with pkg_resources.path(readme_examples, 'create_component_example.simultan') as r_path:
    project_path = str(r_path)

data_model = DataModel.create_new_project(project_path=project_path,
                                          user_name='admin',
                                          password='admin')

# create a new component
new_comp = data_model.create_component(name='Example7 Test Component',
                                       Visibility=1,
                                       IsAutomaticallyGenerated=False)


# create a new component which is going to be the sub component
sub_component = data_model.create_component(name='Example7 New Sub Component',
                                            Visibility=1,
                                            IsAutomaticallyGenerated=False)

# add the sub_component as sub component to new_comp
add_sub_component(new_comp, sub_component, "Undefined Slot", 15)

data_model.save()
data_model.cleanup()



import hashlib
import json

from copy import copy
from weakref import WeakSet
from functools import lru_cache
from . import utils
from numpy import ndarray
from pandas import DataFrame
import colorlog
from typing import Union, Optional, Any

logger = colorlog.getLogger('PySimultan')

from SIMULTAN.Data.Taxonomy import SimTaxonomyEntry, SimTaxonomyEntryReference, SimTaxonomy
from SIMULTAN.Data.Components import (ComponentWalker, SimComponent, SimBoolParameter, SimDoubleParameter,
                                      SimEnumParameter, SimIntegerParameter, SimStringParameter, ComponentMapping,
                                      SimSlot, ComponentMapping)
from SIMULTAN.Data.Assets import ResourceEntry, ResourceFileEntry, ContainedResourceFileEntry, Asset
from SIMULTAN.Data import SimId

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .data_model import DataModel
    from .object_mapper import PythonMapper
    from .geometry.geometry_base import ExtendedBaseGeometry

from .geometry.utils import create_python_geometry

from . import config


class ComplexEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, SimultanObject):
            return obj.to_json()
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


class classproperty(object):

    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


class MetaMock(type):
    def __call__(cls: 'SimultanObject', *args, **kwargs) -> 'SimultanObject':
        """
        Metaclass to implement object initialization either with wrapped_obj or keywords.

        If a wrapped_obj is defined, create new SimultanObject which wraps a SIMULTAN component (wrapped_obj).

        If no 'wrapped_obj' is defined, a new SimComponent is created with the content defined in the template and the
        values are set to the values defined in kwargs.
        """

        obj = cls.__new__(cls, *args, **kwargs)

        wrapped_obj = kwargs.get('wrapped_obj', None)
        if wrapped_obj is None:

            data_model = kwargs.get('data_model', None)
            if data_model is None:
                if config.get_default_data_model() is not None:
                    logger.warning(
                        f'No data model provided. Using default data model: {config.get_default_data_model().id}')
                    data_model = config.get_default_data_model()
                    kwargs['data_model'] = data_model
                else:
                    raise TypeError((f'Error creating new instance of class {cls.__name__}:\n'
                                     f'Any data model was defined. Tried to use default data model but there are multiple datamodels.\n'
                                     f'Define the data model to use with the key: data_model'))

            wrapped_obj = cls.create_simultan_component(*args, **kwargs)

            init_dict = kwargs.copy()
            init_dict['data_model_id'] = data_model.id
            init_dict['object_mapper'] = cls._object_mapper
            init_dict['data_model'] = data_model
            init_dict['wrapped_obj'] = wrapped_obj

            obj.__init__(*args, **init_dict)

            # for key, value in kwargs.items():
            #     if key in ['data_model', 'wrapped_obj', 'object_mapper']:
            #         continue
            #     setattr(obj, key, value)

        else:
            obj.__init__(*args, **kwargs)
        return obj


class SimultanObject(object, metaclass=MetaMock):

    _cls_instances = set()  # weak set with all created objects
    _create_all = False  # if true all properties are evaluated to create python objects when initialized
    _cls_instances_dict_cache = None
    __type_view__ = None      # ui view cls

    @classmethod
    def reset_cls(cls):
        cls._cls_instances = set()
        cls._cls_instances_dict_cache = None

    @classmethod
    def get_instance_by_id(cls, id: SimId) -> 'SimultanObject':
        return cls._cls_instances_dict.get(id, None)

    @classproperty
    def _cls_instances_dict(cls) -> dict[SimId, 'SimultanObject']:
        if cls._cls_instances_dict_cache is None:
            cls._cls_instances_dict_cache = dict(zip([x.id for x in cls._cls_instances], [x for x in cls._cls_instances]))

        elif len(cls._cls_instances) == 0:
            return {}
        elif len(cls._cls_instances) != len(cls._cls_instances_dict_cache):
            cls._cls_instances_dict_cache = dict(zip([x.id for x in cls._cls_instances], [x for x in cls._cls_instances]))
        return cls._cls_instances_dict_cache

    @classproperty
    def cls_instances(cls) -> list['SimultanObject']:
        try:
            return list(cls._cls_instances)
        except Exception as e:
            logger.error(f'Error getting cls_instances: {e}')
            return []

    @classmethod
    def create_simultan_component(cls, *args, **kwargs) -> SimComponent:
        wrapped_obj = utils.create_simultan_component_for_taxonomy(cls, *args, **kwargs)
        wrapped_obj.Name = kwargs.get('name', 'UnnamedComponent')
        return wrapped_obj

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        if "_cls_instances" not in cls.__dict__:
            cls._cls_instances = set()
        try:
            cls._cls_instances.add(instance)
        except Exception as e:
            logger.error(f'Error adding instance {instance} to _cls_instances: {e}')

        return instance

    def __init__(self, *args, **kwargs):
        """
        Initialize the SimultanObject. This method is only called if the object is initialized the first time without a
        wrapped_obj. If a wrapped_obj is defined, the __new__ method is called and the object can be initialized with
        the __load_init__ method.

        :param args:
        :param kwargs:
        """

        self._wrapped_obj: Union[SimComponent, None] = kwargs.get('wrapped_obj', None)
        self.__obj_init__: bool = kwargs.get('__obj_init__', False)
        self._data_model: Union[DataModel, None] = kwargs.get('data_model', config.get_default_data_model())

        object_mapper = kwargs.get('object_mapper', None)
        if object_mapper is None:
            object_mapper = config.get_default_mapper()
        self._object_mapper: Union[PythonMapper, None] = object_mapper
        self.name: str = kwargs.get('name', '')

        self.__property_cache__ = {}
        self._slot = None

    def __getattribute__(self, attr) -> Any:

        try:
            return object.__getattribute__(self, attr)
        except AttributeError:
            wrapped = object.__getattribute__(self, '_wrapped_obj')
            if wrapped is not None:
                return object.__getattribute__(wrapped, attr)
            else:
                raise KeyError

    def __setattr__(self, attr, value) -> None:
        if hasattr(self, '_wrapped_obj'):
            if hasattr(self._wrapped_obj, attr) and (self._wrapped_obj is not None) and not self.__obj_init__:
                object.__setattr__(self._wrapped_obj, attr, value)
                return
        # object.__setattr__(self, attr, value)
        super().__setattr__(attr, value)

        # if hasattr(super(self.__class__), attr):
        #     super(self.__class__).__setattr__(attr, value)

    @property
    def id(self) -> SimId:
        if self._wrapped_obj is not None:
            return self._wrapped_obj.Id

    @property
    def id_str(self) -> str:
        return '_'.join((str(self.id.GlobalId), str(self.id.LocalId)))

    @property
    def taxonomy_keys(self) -> list[str]:
        if self._wrapped_obj is not None:
            return [x.Target.Key for x in self._wrapped_obj.Slots]

    @property
    def name(self) -> str:
        if self._wrapped_obj is not None:
            return self._wrapped_obj.Name

    @name.setter
    def name(self, value: str):
        if self._wrapped_obj is not None:
            self._wrapped_obj.Name = value

    @property
    def parent(self) -> 'SimultanObject':
        if not hasattr(self._wrapped_obj, 'Parent'):
            return None

        if self._wrapped_obj.Parent is not None:
            return self._object_mapper.create_python_object(self._wrapped_obj.Parent, data_model=self._data_model)
        else:
            return None

    @property
    def slots(self):
        if self._wrapped_obj is not None:
            return list(self._wrapped_obj.Slots.Items)

    @property
    def primary_slot(self):
        if self._wrapped_obj is not None:
            return self._wrapped_obj.Slots.Items[0]

    @property
    def referenced_by(self) -> set['SimultanObject']:
        return set([self._object_mapper.create_python_object(x.Target, data_model=self._data_model)
                    for x in self._wrapped_obj.ReferencedBy if
                    x.Target != self._wrapped_obj])

    @property
    def _parameters(self) -> dict[str, Union[str, int, float, bool, ndarray]]:
        return {x.NameTaxonomyEntry.get_TextOrKey(): utils.get_obj_value(x,
                                                                         data_model=self._data_model,
                                                                         object_mapper=self._object_mapper) for x in
                self._wrapped_obj.Parameters}

    @property
    def associated_geometry(self):
        ref_geometries = self._data_model.get_associated_geometry(self)

        return [create_python_geometry(None, geo, self._data_model, self._object_mapper, geo_model) for geo, geo_model in
                ref_geometries]

    def get_subcomponents(self) -> dict[[str, str], Union['SimultanObject', SimComponent]]:
        subcomponents = {}
        for comp in self._wrapped_obj.Components:
            subcomponents[(comp.Slot.SlotBase.Target.Key,
                           comp.Slot.SlotExtension)] = self._object_mapper.create_python_object(comp.Component,
                                                                                                data_model=self._data_model)
        return subcomponents

    def get_subcomponent_list(self) -> list[Union['SimultanObject', SimComponent]]:
        return list(self.get_subcomponents().values())

    def get_referenced_components(self) -> list['SimultanObject']:
        return [self._object_mapper.create_python_object(x, data_model=self._data_model)
                for x in self._wrapped_obj.ReferencedComponents]

    def add_taxonomy_entry_reference(self,
                                     taxonomy_entry_reference: SimTaxonomyEntryReference,
                                     index: int = None):
        if index is None:
            index = self._wrapped_obj.Slots.Items.__len__()
        self._wrapped_obj.Slots.InsertItem(index, taxonomy_entry_reference)

    def add_subcomponent(self, subcomponent: Union['SimultanObject', SimComponent],
                         slot: SimTaxonomyEntryReference = None,
                         slot_extension: str = None) -> 'SimultanObject':

        if isinstance(subcomponent, SimComponent):
            comp_to_add = subcomponent
        elif isinstance(subcomponent, SimultanObject):
            comp_to_add = subcomponent._wrapped_obj
        else:
            comp_to_add = utils.create_mapped_python_object(subcomponent,
                                                            object_mapper=self._object_mapper,
                                                            data_model=self._data_model,
                                                            add_to_data_model=True)

        utils.add_sub_component(self._wrapped_obj,
                                comp_to_add,
                                slot_extension,
                                slot)

        return comp_to_add

    def remove_subcomponent(self, subcomponent: Union['SimultanObject', SimComponent]):
        if isinstance(subcomponent, SimComponent):
            utils.remove_sub_component(self._wrapped_obj, subcomponent)
        elif isinstance(subcomponent, SimultanObject):
            utils.remove_sub_component(self._wrapped_obj, subcomponent._wrapped_obj)
        else:
            raise TypeError(f'Unknown type for subcomponent: {type(subcomponent)}')

    def add_referenced_component(self,
                                 referenced_component: Union['SimultanObject', SimComponent],
                                 slot: SimTaxonomyEntryReference = None,
                                 slot_extension: str = None):

        referenced_wrapped_obj = referenced_component if isinstance(referenced_component,
                                                                    SimComponent) else referenced_component._wrapped_obj

        # logger.debug(f'Adding referenced component {referenced_component} to {self}')
        utils.add_referenced_component(self._wrapped_obj,
                                       referenced_wrapped_obj,
                                       slot_extension,
                                       slot)

    def remove_referenced_component(self, referenced_component: 'SimultanObject'):

        if isinstance(referenced_component, SimComponent):
            utils.remove_referenced_component(self._wrapped_obj, referenced_component)
        elif isinstance(referenced_component, SimultanObject):
            utils.remove_referenced_component(self._wrapped_obj, referenced_component._wrapped_obj)


    def add_asset(self,
                  resource_file_entry: 'ResourceFileEntry',
                  _id_contained: str = '') -> Asset:

        return ComponentMapping.AddAsset(self, resource_file_entry, _id_contained)

    def __repr__(self):
        return f'{self.name}: ' + object.__repr__(self)

    def associate(self, other: 'ExtendedBaseGeometry'):
        """
        Associate this object with another object
        :param other: fc_geometry object to associate with
        :return: None
        """
        other.associate(self)

    def get_raw_attr(self, attr: Optional[str] = None, text_or_key: Optional[str] = None):
        if attr is not None:
            content = self._taxonomy_map.get_content_by_property_name(attr)
            if content is None:
                raise KeyError(f'No content found for attribute {attr}')

            return utils.get_component_taxonomy_entry(self._wrapped_obj, content.text_or_key)

        if text_or_key is not None:
            return utils.get_component_taxonomy_entry(self._wrapped_obj, text_or_key)

    def set_attr_prop(self, attr: str, prop: str, value):
        """
        Set attribute of the mapped property

        Example:

        class TestComponent(object):
            def __init__(self, *args, **kwargs):
                self.value = kwargs.get('value')

        content0 = Content(text_or_key='value',  # text or key of the content/parameter/property
                           property_name='value',  # name of the generated property
                           type=float,  # type of the content/parameter/property
                           unit=None,  # unit of the content/parameter/property
                           documentation='value to test',
                           component_policy='subcomponent')

        test_component_map = TaxonomyMap(taxonomy_name='PySimultan',
                                         taxonomy_key='PySimultan',
                                         taxonomy_entry_name='TestComponent',
                                         taxonomy_entry_key='TestComponent',
                                         content=[content0]
                                         )

        mapper.register(test_component_map.taxonomy_entry_key, TestComponent, taxonomy_map=test_component_map)
        cls1 = mapper.get_mapped_class(test_component_map.taxonomy_entry_key)

        test_component = cls1(name='test_component',
                                value=None)

        test_component.set_attr_prop('value', 'MinValue, 0)
        test_component.set_attr_prop('value', 'MaxValue, 1)

        :param attr:
        :param value:
        :return:
        """
        comp_prop = utils.get_component_taxonomy_entry(self._wrapped_obj, attr)
        if comp_prop is not None:
            setattr(comp_prop, prop, value)
        return comp_prop

    def copy(self,
             name: str = None):
        """
        Create a copy of the object including all properties. The properties are copied by reference.
        :return:
        """
        new_copy = self.__class__(name=name if name is not None else f'Copy of {self.name}',
                                  data_model=self._data_model,
                                  object_mapper=self._object_mapper)

        for content in self._taxonomy_map.content:
            setattr(new_copy, content.property_name, getattr(self, content.property_name))

        # new_copy.name = name if name is not None else f'Copy of {self.name}'

        if self.associated_geometry:
            for geo in self.associated_geometry:
                new_copy.associate(geo)

        return new_copy

    def remove_from_datamodel(self):
        self._data_model.remove_component(self)
        self.__class__.cls_instances.remove(self)

    def to_json(self) -> dict:

        from .default_types import ComponentList, ComponentDictionary

        obj_dict = {
            str(self.id): {'name': self.name},
            'taxonomies': self.taxonomy_keys,
                    }

        for content in self._taxonomy_map.content:
            val = getattr(self, content.property_name)

            if hasattr(val, 'json_ref'):
                obj_dict[content.text_or_key] = val.json_ref()
            elif isinstance(val, (list, ComponentList)):
                obj_dict[content.text_or_key] = [v.json_ref() if hasattr(v, 'json_ref') else v for v in val]
            elif isinstance(val, (dict, ComponentDictionary)):
                obj_dict[content.text_or_key] = {k: v.json_ref() if hasattr(v, 'json_ref') else v for k, v in val.items()}
            elif isinstance(val, ndarray):
                obj_dict[content.text_or_key] = val.tolist()
            elif isinstance(val, DataFrame):
                obj_dict[content.text_or_key] = val.to_dict()
            else:
                obj_dict[content.text_or_key] = val

        return obj_dict

    def json_ref(self):
        return {"$ref": {
            "$type": 'Component',
            "$taxonomies": self.taxonomy_keys,
            "$id": {'local_id': self.id.LocalId,
                    'global_id': str(self.id.GlobalId)
                    }
        }
        }

    # def to_json(self):
    #
    #     super().to_json(self)

        # return {str(self.id): {'name': self.name}}



from PySimultan2.data_model import DataModel

project_path = r'resources/U5.simultan'

data_model = DataModel(project_path=project_path)

# data_model.data.Items[3].Components[0].Component.Slots[0].Target.Key

of_solids = data_model.find_components_with_taxonomy(taxonomy='of_solid')
of_volumes = data_model.find_components_with_taxonomy(taxonomy='of_volume')

# list(list(surface_mesh_setups)[0].Parameters)[0].get_NameTaxonomyEntry().TextOrKey
# list(list(surface_mesh_setups)[0].Parameters)[0].NameTaxonomyEntry

ref = data_model.get_associated_geometry(list(of_volumes)[0])

c = data_model.get_referenced_components(ref[0])

# fc_geometry = list(data_model.models)[2].Geometry.GeometryFromId(367)
# component = data_model.exch.GetComponents(fc_geometry)

[list(data_model.exch.GetComponents(x)) for x in list(list(data_model.models.values())[2].Geometry.Volumes)]

print('ok')



import os
from time import sleep

from numpy import isclose
from PySimultan2.src.PySimultan2 import DataModel
from PySimultan2.src.PySimultan2 import PythonMapper
from PySimultan2.src.PySimultan2.taxonomy_maps import TaxonomyMap, Content

from typing import List

from PySimultan2.tests import resources
from PySimultan2.src.PySimultan2.geometry.geometry_base import (GeometryModel)
from PySimultan2.src.PySimultan2.geometry.utils import create_cube

try:
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources


project_path = os.environ.get('SIMULTAN_PROJECT_PATH', None)

if project_path is None:
    with pkg_resources.path(resources, 'new_geometry_test.simultan') as r_path:
        project_path = str(r_path)
else:
    project_path = os.path.join(project_path, 'new_geometry_test.simultan')


def create_geometry_model(name='new_geometry_test', data_model=None):
    return GeometryModel(name=name,
                         data_model=data_model)


def create_classes() -> tuple[dict[str, type], PythonMapper]:

    class Source(object):
        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name')
            self.temperature = kwargs.get('temperature')
            self.heat_flux = kwargs.get('heat_flux')

    class TemperatureBoundaryCondition(object):
        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name')
            self.temperature = kwargs.get('temperature')

    class HeatFluxBoundaryCondition(object):
        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name')
            self.heat_flux = kwargs.get('heat_flux')

    class Material(object):
        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name')
            self.thermal_conductivity = kwargs.get('thermal_conductivity')
            self.density = kwargs.get('density')
            self.specific_heat = kwargs.get('specific_heat')

    class WallLayer(object):
        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name')
            self.thickness = kwargs.get('thickness')
            self.material: Material = kwargs.get('material')

        @property
        def r_value(self):
            return self.thickness / self.material.thermal_conductivity
        
    class Construction(object):
        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name')
            self.layers: List[WallLayer] = kwargs.get('layers', [])
            self.r_value = kwargs.get('r_value', None)

        def calculate_r_value(self):
            self.r_value = sum([layer.r_value for layer in self.layers])
            return self.r_value

    class Wall(object):

        h_c_lookup = {0.1: 5,
                      0.13: 7.69,
                      0.17: 5.88}

        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name')
            self.r_si = kwargs.get('r_si')
            self.r_se = kwargs.get('r_se')
            self.construction: Construction = kwargs.get('construction')
            self.u_value = kwargs.get('u_value', None)

            self.boundary_condition = kwargs.get('boundary_condition', None)

        @property
        def area(self):

            if not self.associated_geometry:
                raise ValueError('No associated fc_geometry')

            return sum((x.area for x in self.associated_geometry))

        @property
        def h_c(self):
            """
            return the convective heat transfer coefficient h_c in W/m2K
            :return:
            """
            return self.h_c_lookup[self.r_si]

        def calculate_u_value(self) -> float:
            self.u_value = 1 / (self.r_si + self.r_se + self.construction.calculate_r_value())
            return self.u_value

    class Zone(object):
        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name')
            self.walls: List[Wall] = kwargs.get('walls', [])
            self.volume = kwargs.get('volume')
            self.area = kwargs.get('area')
            self.u_value = kwargs.get('u_value', None)

            self.sources: List[Source] = kwargs.get('sources', [])
            self.steady_state_temperature = kwargs.get('steady_state_temperature', None)

        def calculate_steady_state_temperature(self):
            try:
                phi_ti = []

                print('Pos 4.1: ' + str(self._data_model.project_data_manager.AssetManager.Resources.Count) +
                      f'Project manager id: {id(self._data_model._project_data_manager)} \n'
                      f'Data Model id: {id(self._data_model)} \n')

                for wall in self.walls:
                    phi_i = wall.area * wall.calculate_u_value() * wall.r_si * wall.h_c
                    phi_ti.append((phi_i * wall.boundary_condition.temperature, phi_i))

                phi_hc_id = sum(x.heat_flux for x in self.sources)

                t_steady = (sum(x[0] for x in phi_ti) + phi_hc_id) / sum(x[1] for x in phi_ti)
                self.steady_state_temperature = t_steady
            except Exception as e:
                raise e

            return t_steady

    cls_dict = {'Material': Material,
                'WallLayer': WallLayer,
                'Wall': Wall,
                'Zone': Zone,
                'TemperatureBoundaryCondition': TemperatureBoundaryCondition,
                'HeatFluxBoundaryCondition': HeatFluxBoundaryCondition,
                'Source': Source,
                'Construction': Construction}

    return cls_dict


def create_mapped_classes(classes: dict[str, type]) -> dict[str, type]:

    mapper = PythonMapper()
    
    def create_contents() -> dict[str, Content]:
        
        contents = {}

        contents['thermal_conductivity'] = Content(text_or_key='thermal_conductivity',
                                                   property_name='thermal_conductivity',
                                                   type=None,
                                                   unit='W/mK',
                                                   documentation='thermal_conductivity in W/mK')

        contents['density'] = Content(text_or_key='density',  
                                      property_name='density', 
                                      type=None,  
                                      unit='kg/m3',
                                      documentation='density in kg/m3')

        contents['specific_heat'] = Content(text_or_key='specific_heat',  
                                            property_name='specific_heat', 
                                            type=None,  
                                            unit='J/kgK',
                                            documentation='specific_heat in J/kgK')

        contents['thickness'] = Content(text_or_key='thickness',  
                                        property_name='thickness', 
                                        type=None,  
                                        unit='m',
                                        documentation='thickness in m')

        contents['material'] = Content(text_or_key='material',  
                                       property_name='material', 
                                       type=None,  
                                       unit=None,  
                                       documentation='material')

        contents['construction'] = Content(text_or_key='construction',
                                           property_name='construction',
                                           type=None,
                                           unit=None,
                                           documentation='construction')

        contents['r_si'] = Content(text_or_key='r_si',  
                                   property_name='r_si', 
                                   type=None,  
                                   unit='m2K/W',
                                   documentation='r_si in m2K/W')

        contents['r_se'] = Content(text_or_key='r_se',  
                                   property_name='r_se', 
                                   type=None,  
                                   unit=None,  
                                   documentation='r_se in m2K/W')

        contents['layers'] = Content(text_or_key='layers',  
                                     property_name='layers', 
                                     type=None,  
                                     unit=None,  
                                     documentation='layers of the wall',
                                     component_policy='subcomponent')

        contents['u_value'] = Content(text_or_key='u_value',  
                                      property_name='u_value', 
                                      type=None,  
                                      unit='W/m2K',
                                      documentation='u_value of the wall in W/m2K')

        contents['r_value'] = Content(text_or_key='r_value',
                                      property_name='r_value',
                                      type=None,
                                      unit='m2K/W',
                                      documentation='r_value of the wall in W/m2K')

        contents['boundary_condition'] = Content(text_or_key='boundary_condition',  
                                                 property_name='boundary_condition', 
                                                 type=None,  
                                                 unit=None,  
                                                 documentation='boundary_condition')

        contents['walls'] = Content(text_or_key='simultan_walls',
                                    property_name='walls',
                                    type=None,
                                    unit=None,
                                    documentation='walls of the zone')

        contents['volume'] = Content(text_or_key='volume',
                                     property_name='volume',
                                     type=None,
                                     unit=None,
                                     documentation='volume of the zone in m3')

        contents['area'] = Content(text_or_key='area',
                                   property_name='area',
                                   type=None,
                                   unit='m2',
                                   documentation='area of the zone in m2')

        contents['sources'] = Content(text_or_key='sources',
                                      property_name='sources',
                                      type=None,
                                      unit=None,
                                      documentation='sources of the zone')

        contents['steady_state_temperature'] = Content(text_or_key='steady_state_temperature',
                                                       property_name='steady_state_temperature',
                                                       type=None,
                                                       unit='K',
                                                       documentation='steady_state_temperature of the zone')

        contents['temperature'] = Content(text_or_key='temperature',
                                          property_name='temperature',
                                          type=None,
                                          unit='K',
                                          documentation='temperature in K')

        contents['heat_flux'] = Content(text_or_key='heat_flux',
                                        property_name='heat_flux',
                                        type=None,
                                        unit='W',
                                        documentation='heat_flux in W')

        return contents

    def create_mapped_material(cls, contents: dict[str, Content]):

        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='Material',
                              taxonomy_entry_key='Material',
                              content=[contents['thermal_conductivity'],
                                       contents['density'],
                                       contents['specific_heat']],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    def create_layer_cls(cls, contents: dict[str, Content]):

        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='Layer',
                              taxonomy_entry_key='Layer',
                              content=[contents['thickness'], contents['material']],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    def create_construction_cls(cls, contents: dict[str, Content]):

        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='Construction',
                              taxonomy_entry_key='Construction',
                              content=[contents['layers'],
                                       contents['r_value']
                                       ],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    def create_wall_cls(cls, contents: dict[str, Content]):

        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='Wall',
                              taxonomy_entry_key='Wall',
                              content=[contents['r_si'],
                                       contents['r_se'],
                                       contents['construction'],
                                       contents['u_value'],
                                       contents['boundary_condition']],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    def create_zone_cls(cls, contents: dict[str, Content]):

        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='Zone',
                              taxonomy_entry_key='Zone',
                              content=[contents['walls'],
                                       contents['volume'],
                                       contents['area'],
                                       contents['u_value'],
                                       contents['sources'],
                                       contents['steady_state_temperature']],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    def create_temperature_bc_cls(cls, contents: dict[str, Content]):
        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='TemperatureBoundaryCondition',
                              taxonomy_entry_key='TemperatureBoundaryCondition',
                              content=[contents['temperature']],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    def create_heat_flux_bc_cls(cls, contents: dict[str, Content]):
        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='HeatFluxBoundaryCondition',
                              taxonomy_entry_key='HeatFluxBoundaryCondition',
                              content=[contents['heat_flux']],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    def create_source_cls(cls, contents: dict[str, Content]):
        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='Source',
                              taxonomy_entry_key='Source',
                              content=[contents['temperature'],
                                       contents['heat_flux']],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    tax_contents = create_contents()

    mapped_material_cls = create_mapped_material(classes['Material'], tax_contents)
    mapped_layer_cls = create_layer_cls(classes['WallLayer'], tax_contents)
    mapped_construction_cls = create_construction_cls(classes['Construction'], tax_contents)
    mapped_wall_cls = create_wall_cls(classes['Wall'], tax_contents)
    mapped_zone_cls = create_zone_cls(classes['Zone'], tax_contents)
    mapped_temperature_bc_cls = create_temperature_bc_cls(classes['TemperatureBoundaryCondition'], tax_contents)
    mapped_heat_flux_bc_cls = create_heat_flux_bc_cls(classes['HeatFluxBoundaryCondition'], tax_contents)
    mapped_source_cls = create_source_cls(classes['Source'], tax_contents)

    mapped_cls_dict = {'Material': mapped_material_cls,
                       'Layer': mapped_layer_cls,
                       'Construction': mapped_construction_cls,
                       'Wall': mapped_wall_cls,
                       'Zone': mapped_zone_cls,
                       'TemperatureBoundaryCondition': mapped_temperature_bc_cls,
                       'HeatFluxBoundaryCondition': mapped_heat_flux_bc_cls,
                       'Source': mapped_source_cls}

    return mapped_cls_dict, mapper


def create_model(project_name=None):

    if project_name is None:
        project_name = project_path

    data_model = DataModel.create_new_project(project_path=project_name,
                                              user_name='admin',
                                              password='admin')

    geo_model = create_geometry_model(name='new_geometry_test',
                                      data_model=data_model)

    cube = create_cube(data_model, geo_model, scale=10)

    classes = create_classes()
    mapped_classes, mapper = create_mapped_classes(classes)

    Material = mapped_classes['Material']
    WallLayer = mapped_classes['Layer']
    Construction = mapped_classes['Construction']
    Wall = mapped_classes['Wall']
    Zone = mapped_classes['Zone']
    TemperatureBoundaryCondition = mapped_classes['TemperatureBoundaryCondition']
    HeatFluxBoundaryCondition = mapped_classes['HeatFluxBoundaryCondition']
    Source = mapped_classes['Source']

    t_out_1 = TemperatureBoundaryCondition(name='t_out_1',
                                           temperature=273.15)

    t_out_2 = TemperatureBoundaryCondition(name='t_out_2',
                                           temperature=283.15)

    concrete = Material(name='concrete',
                        thermal_conductivity=1.5,
                        density=2000,
                        specific_heat=800)

    insulation = Material(name='insulation',
                          thermal_conductivity=0.03,
                          density=150,
                          specific_heat=1500)

    plaster = Material(name='plaster',
                       thermal_conductivity=0.6,
                       density=1200,
                       specific_heat=700)

    construction_1 = Construction(name='construction_1',
                                  layers=[WallLayer(name='concrete_layer', thickness=0.2, material=concrete),
                                          WallLayer(name='insulation_layer', thickness=0.1, material=insulation),
                                          WallLayer(name='plaster_layer', thickness=0.01, material=plaster)])

    construction_2 = Construction(name='construction_2',
                                  layers=[WallLayer(name='concrete_layer', thickness=0.15, material=concrete),
                                          WallLayer(name='insulation_layer', thickness=0.2, material=insulation)
                                          ]
                                  )

    wall1 = Wall(name='Wall1',
                 r_si=0.13,
                 r_se=0.04,
                 construction=construction_1,
                 boundary_condition=t_out_1)
    wall1.associate(cube.faces[0])

    _ = wall1.construction
    _ = wall1.construction
    wall1.construction = construction_2
    _ = wall1.construction
    wall1.construction = construction_1
    _ = wall1.construction

    _ = wall1.associated_geometry

    wall2 = Wall(name='Wall2',
                 r_si=0.13,
                 r_se=0.04,
                 construction=construction_2,
                 boundary_condition=t_out_2)
    wall2.associate(cube.faces[1])

    _ = wall2.construction

    wall3 = Wall(name='Wall3',
                 r_si=0.13,
                 r_se=0.04,
                 construction=construction_1,
                 boundary_condition=t_out_1)
    wall3.associate(cube.faces[2])

    _ = wall3.construction

    wall4 = Wall(name='Wall4',
                 r_si=0.13,
                 r_se=0.04,
                 construction=construction_2,
                 boundary_condition=t_out_2)
    wall4.associate(cube.faces[3])

    wall5 = Wall(name='Wall5',
                 r_si=0.13,
                 r_se=0.04,
                 construction=construction_2,
                 boundary_condition=t_out_1)
    wall5.associate(cube.faces[4])

    wall6 = Wall(name='Wall6',
                 r_si=0.13,
                 r_se=0.04,
                 construction=construction_2,
                 boundary_condition=t_out_2)
    wall6.associate(cube.faces[5])

    convective_heat_source = Source(name='convective_heat_source',
                                    heat_flux=1000)

    zone1 = Zone(name='zone_1',
                 walls=[wall1, wall2, wall3, wall4, wall5, wall6],
                 sources=[convective_heat_source])
    zone1.associate(cube)

    print(cube.components)
    print(list(zone1.associated_geometry))

    steady_state_temperature = zone1.calculate_steady_state_temperature()
    assert isclose(steady_state_temperature, 285.8980257275199)
    print(steady_state_temperature)

    data_model.save()

    data_model.cleanup()
    mapper.clear()


def test_loaded_model(project_name=None):

    if project_name is None:
        project_name = project_path

    mapped_cls_dict, mapper = create_mapped_classes(create_classes())

    data_model = DataModel(project_path=project_name,
                           user_name='admin',
                           password='admin')

    # mapper.load_undefined = False
    mapper.load_undefined = True

    typed_data = data_model.get_typed_data(mapper=mapper,
                                           create_all=True)

    new_undef_component = mapper.get_mapped_class('new_tax')()

    zone_cls = mapper.get_mapped_class('Zone')

    zone1 = zone_cls.cls_instances[0]

    temp = zone1.calculate_steady_state_temperature()

    assert isclose(temp, 285.8980257275199)

    data_model.cleanup()
    mapper.clear()
    del data_model


if __name__ == '__main__':

    for i in range(3):
        print(f'Iteration {i}')
        # create_model()
        # sleep(1)
        test_loaded_model()
        sleep(1)



import typing, clr, abc
from System import IDisposable, SystemException, Exception, Attribute, Array_1
from System.Collections import IEnumerable, IDictionaryEnumerator, IDictionary
from System.Reflection import MethodBase, Assembly
from System.Globalization import CultureInfo
from System.IO import UnmanagedMemoryStream, Stream

class IResourceReader(IDisposable, IEnumerable, typing.Protocol):
    @abc.abstractmethod
    def Close(self) -> None: ...
    @abc.abstractmethod
    def GetEnumerator(self) -> IDictionaryEnumerator: ...


class MissingManifestResourceException(SystemException):
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, message: str) -> None: ...
    @typing.overload
    def __init__(self, message: str, inner: Exception) -> None: ...
    @property
    def Data(self) -> IDictionary: ...
    @property
    def HelpLink(self) -> str: ...
    @HelpLink.setter
    def HelpLink(self, value: str) -> str: ...
    @property
    def HResult(self) -> int: ...
    @HResult.setter
    def HResult(self, value: int) -> int: ...
    @property
    def InnerException(self) -> Exception: ...
    @property
    def Message(self) -> str: ...
    @property
    def Source(self) -> str: ...
    @Source.setter
    def Source(self, value: str) -> str: ...
    @property
    def StackTrace(self) -> str: ...
    @property
    def TargetSite(self) -> MethodBase: ...


class MissingSatelliteAssemblyException(SystemException):
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, message: str) -> None: ...
    @typing.overload
    def __init__(self, message: str, cultureName: str) -> None: ...
    @typing.overload
    def __init__(self, message: str, inner: Exception) -> None: ...
    @property
    def CultureName(self) -> str: ...
    @property
    def Data(self) -> IDictionary: ...
    @property
    def HelpLink(self) -> str: ...
    @HelpLink.setter
    def HelpLink(self, value: str) -> str: ...
    @property
    def HResult(self) -> int: ...
    @HResult.setter
    def HResult(self, value: int) -> int: ...
    @property
    def InnerException(self) -> Exception: ...
    @property
    def Message(self) -> str: ...
    @property
    def Source(self) -> str: ...
    @Source.setter
    def Source(self, value: str) -> str: ...
    @property
    def StackTrace(self) -> str: ...
    @property
    def TargetSite(self) -> MethodBase: ...


class NeutralResourcesLanguageAttribute(Attribute):
    @typing.overload
    def __init__(self, cultureName: str) -> None: ...
    @typing.overload
    def __init__(self, cultureName: str, location: UltimateResourceFallbackLocation) -> None: ...
    @property
    def CultureName(self) -> str: ...
    @property
    def Location(self) -> UltimateResourceFallbackLocation: ...
    @property
    def TypeId(self) -> typing.Any: ...


class ResourceManager:
    @typing.overload
    def __init__(self, baseName: str, assembly: Assembly) -> None: ...
    @typing.overload
    def __init__(self, baseName: str, assembly: Assembly, usingResourceSet: typing.Type[typing.Any]) -> None: ...
    @typing.overload
    def __init__(self, resourceSource: typing.Type[typing.Any]) -> None: ...
    HeaderVersionNumber : int
    MagicNumber : int
    @property
    def BaseName(self) -> str: ...
    @property
    def IgnoreCase(self) -> bool: ...
    @IgnoreCase.setter
    def IgnoreCase(self, value: bool) -> bool: ...
    @property
    def ResourceSetType(self) -> typing.Type[typing.Any]: ...
    @staticmethod
    def CreateFileBasedResourceManager(baseName: str, resourceDir: str, usingResourceSet: typing.Type[typing.Any]) -> ResourceManager: ...
    def GetResourceSet(self, culture: CultureInfo, createIfNotExists: bool, tryParents: bool) -> ResourceSet: ...
    def ReleaseAllResources(self) -> None: ...
    # Skipped GetObject due to it being static, abstract and generic.

    GetObject : GetObject_MethodGroup
    class GetObject_MethodGroup:
        @typing.overload
        def __call__(self, name: str) -> typing.Any:...
        @typing.overload
        def __call__(self, name: str, culture: CultureInfo) -> typing.Any:...

    # Skipped GetStream due to it being static, abstract and generic.

    GetStream : GetStream_MethodGroup
    class GetStream_MethodGroup:
        @typing.overload
        def __call__(self, name: str) -> UnmanagedMemoryStream:...
        @typing.overload
        def __call__(self, name: str, culture: CultureInfo) -> UnmanagedMemoryStream:...

    # Skipped GetString due to it being static, abstract and generic.

    GetString : GetString_MethodGroup
    class GetString_MethodGroup:
        @typing.overload
        def __call__(self, name: str) -> str:...
        @typing.overload
        def __call__(self, name: str, culture: CultureInfo) -> str:...



class ResourceReader(IResourceReader):
    @typing.overload
    def __init__(self, fileName: str) -> None: ...
    @typing.overload
    def __init__(self, stream: Stream) -> None: ...
    def Close(self) -> None: ...
    def Dispose(self) -> None: ...
    def GetEnumerator(self) -> IDictionaryEnumerator: ...
    def GetResourceData(self, resourceName: str, resourceType: clr.Reference[str], resourceData: clr.Reference[Array_1[int]]) -> None: ...


class ResourceSet(IEnumerable, IDisposable):
    @typing.overload
    def __init__(self, fileName: str) -> None: ...
    @typing.overload
    def __init__(self, reader: IResourceReader) -> None: ...
    @typing.overload
    def __init__(self, stream: Stream) -> None: ...
    def Close(self) -> None: ...
    def Dispose(self) -> None: ...
    def GetDefaultReader(self) -> typing.Type[typing.Any]: ...
    def GetDefaultWriter(self) -> typing.Type[typing.Any]: ...
    def GetEnumerator(self) -> IDictionaryEnumerator: ...
    # Skipped GetObject due to it being static, abstract and generic.

    GetObject : GetObject_MethodGroup
    class GetObject_MethodGroup:
        @typing.overload
        def __call__(self, name: str) -> typing.Any:...
        @typing.overload
        def __call__(self, name: str, ignoreCase: bool) -> typing.Any:...

    # Skipped GetString due to it being static, abstract and generic.

    GetString : GetString_MethodGroup
    class GetString_MethodGroup:
        @typing.overload
        def __call__(self, name: str) -> str:...
        @typing.overload
        def __call__(self, name: str, ignoreCase: bool) -> str:...



class SatelliteContractVersionAttribute(Attribute):
    def __init__(self, version: str) -> None: ...
    @property
    def TypeId(self) -> typing.Any: ...
    @property
    def Version(self) -> str: ...


class UltimateResourceFallbackLocation(typing.SupportsInt):
    @typing.overload
    def __init__(self, value : int) -> None: ...
    @typing.overload
    def __init__(self, value : int, force_if_true: bool) -> None: ...
    def __int__(self) -> int: ...
    
    # Values:
    MainAssembly : UltimateResourceFallbackLocation # 0
    Satellite : UltimateResourceFallbackLocation # 1

